MADHUSUDANA_MASA = 0
TRIVIKRAMA_MASA = 1
VAMANA_MASA = 2
SRIDHARA_MASA = 3
HRSIKESA_MASA = 4
PADMANABHA_MASA = 5
DAMODARA_MASA = 6
KESAVA_MASA = 7
NARAYANA_MASA = 8
MADHAVA_MASA = 9
GOVINDA_MASA = 10
VISNU_MASA = 11
ADHIKA_MASA = 12



PRIO_MAHADVADASI = 10
PRIO_EKADASI = 20
PRIO_EKADASI_PARANA = 90
PRIO_FESTIVALS_0 = 100
PRIO_FESTIVALS_1 = 200
PRIO_FESTIVALS_2 = 300
PRIO_FESTIVALS_3 = 400
PRIO_FESTIVALS_4 = 500
PRIO_FESTIVALS_5 = 600
PRIO_FESTIVALS_6 = 700
PRIO_FASTING = 900
PRIO_SANKRANTI = 920
PRIO_MASA_CHANGE = 940
PRIO_DST_CHANGE = 950
PRIO_KSAYA = 965
PRIO_CM_CONT = 971
PRIO_CM_DAY = 972
PRIO_CM_DAYNOTE = 973
PRIO_ARUN = 975
PRIO_SUN  = 980
PRIO_MOON = 990
PRIO_ASTRO = 1000



from .GCGregorianDate import GCGregorianDate, Today
from . import GCEarthData as GCEarthData
from . import GCLocation as GCLocation
from .GCMoonData import MOONDATA
from . import GCAyanamsha as GCAyanamsha
from . import GCStrings as GCStrings
from . import GCMath as GCMath
from .GCTime import GCTime
from .GCSunData import SUNDATA
from . import GCUT as GCUT

import math

def CalculateMidnightNaksatra(date,earth):
    moon = MOONDATA()
    date = GCGregorianDate(date = date)
    date.shour = 1.0
    jdate = date.GetJulianDetailed()
    moon.Calculate(jdate, earth)
    d = GCMath.putIn360( moon.longitude_deg - GCAyanamsha.GetAyanamsa(jdate))
    return math.floor(( d * 3.0) / 40.0 )


def GetNextNaksatra(ed, startDate, nextDate, forward=True):
    phi = 40.0/3.0
    jday = startDate.GetJulianComplete()
    moon = MOONDATA()
    d = GCGregorianDate(date = startDate)
    ayanamsa = GCAyanamsha.GetAyanamsa(jday)
    prev_naks = 0
    new_naks = -1
    dir = 1.0 if forward else -1.0
    scan_step = 0.5 * dir

    xd = GCGregorianDate()

    moon.Calculate(jday, ed)
    l1 = GCMath.putIn360(moon.longitude_deg - ayanamsa)
    prev_naks = int(math.floor(l1 / phi))

    counter = 0
    while counter < 20:
        xj = jday
        xd.Set(d)

        jday += scan_step
        d.shour += scan_step
        d.NormalizeHours()

        moon.Calculate(jday, ed)
        l2 = GCMath.putIn360(moon.longitude_deg - ayanamsa)
        new_naks = int(math.floor(l2/phi))
        if prev_naks != new_naks:
            jday = xj
            d.Set(xd)
            scan_step *= 0.5
            counter+=1
            continue
        else:
            l1 = l2
    nextDate.Set(d)
    return new_naks

def GetPrevNaksatra(ed, startDate, nextDate):
    return GetNextNaksatra(ed, startDate, nextDate, forward=False)

def writeXml(xml, loc, vc, nDaysCount):
    date = GCGregorianDate()

    xml.write("<xml>")
    xml.write("\t<request name=\"Naksatra\" version=\"")
    xml.write(GCStrings.getString(130))
    xml.write("\">\n")
    xml.write("\t\t<arg name=\"longitude\" val=\"")
    xml.write(str(loc.m_fLongitude))
    xml.write("\" />\n")
    xml.write("\t\t<arg name=\"latitude\" val=\"")
    xml.write(str(loc.m_fLatitude))
    xml.write("\" />\n")
    xml.write("\t\t<arg name=\"timezone\" val=\"")
    xml.write(str(loc.m_fTimezone))
    xml.write("\" />\n")
    xml.write("\t\t<arg name=\"startdate\" val=\"")
    xml.write(str(vc))
    xml.write("\" />\n")
    xml.write("\t\t<arg name=\"daycount\" val=\"")
    xml.write(str(nDaysCount))
    xml.write("\" />\n")
    xml.write("\t</request>\n")
    xml.write("\t<result name=\"Naksatra\">\n")

    d = GCGregorianDate(date = vc)
    d.tzone = loc.m_fTimezone
    dn = GCGregorianDate()
    dt = GCTime()
    sun = SUNDATA()
    earth = loc.GetEarthData()

    for i in range(30):
        nak = GetNextNaksatra(earth, d, dn)
        d.Set(dn)
        xml.write("\t\t<day date=\"")
        xml.write(str(d).strip())
        xml.write("\">\n")
        xml.write("\t\t\t<naksatra id=\"")
        xml.write(str(nak))
        xml.write("\" name=\"")
        xml.write(GCStrings.GetNaksatraName(nak))
        xml.write("\" ")
        dt.SetDegTime( d.shour * 360)
        xml.write("starttime=\"")
        xml.write(repr(dt))
        xml.write("\" />\n")

        sun.SunCalc(d, earth);
        xml.write("\t\t\t<sunrise time=\"");
        xml.write(repr(sun.rise))
        xml.write("\" />\n")

        xml.write("\t\t</day>\n");
        d.Set(dn)
        d.shour += 1.0/8.0


    xml.write("\t</result>\n")
    xml.write("</xml>\n")
    return 1

def GetEndHour(earth, yesterday, today):
    nend = GCGregorianDate()
    snd = GCGregorianDate(date = yesterday)
    snd.shour = 0.5
    GetNextNaksatra(earth, snd, nend)
    return nend.GetJulian() - today.GetJulian() + nend.shour

def unittests():
    GCUT.info('naksatra')
    e = GCEarthData.EARTHDATA()
    e.longitude_deg = 27.0
    e.latitude_deg = 45.0
    e.tzone = 1.0
    vc = Today()
    vc2 = GCGregorianDate()
    vc3 = GCGregorianDate(date = vc)
    vc3.NextDay()
    n = GetEndHour(e,vc,vc3)
    GCUT.msg('End hour: {}, {}'.format(n,repr(vc3)))
    import io
    s = io.StringIO()
    clr = GCLocation.GCLocation()
    writeXml(s, clr, vc, 10)
    GCUT.msg(s.getvalue())



from .GCGregorianDate import GCGregorianDate,Today
from .GCEarthData import EARTHDATA
from .GCPancangaDate import GCGaurabdaDate
from .GCLocation import GCLocation
from .GCDayData import GCDayData,GetFirstDayOfYear
from .GCEnums import MasaId
from . import GCTithi as GCTithi
from . import GCStrings as GCStrings
from math import floor
from . import GCUT as GCUT

def GetGaurabdaYear(vc, earth):
    day = GCDayData()
    day.DayCalc(vc, earth)
    day.MasaCalc(vc, earth)
    return day.nGaurabdaYear


def FormatDate(vc,va):
    return "{} {} {}\r\n{}, {} Paksa, {} Masa, {}".format(vc.day, GCStrings.GetMonthAbreviation(vc.month), vc.year, GCStrings.GetTithiName(va.tithi%15), GCStrings.GetPaksaName(va.paksa), GCStrings.GetMasaName(va.masa), va.gyear)

#===========================================================================
#
#===========================================================================

def Gaurabda2Gregorian(va, vc, earth):
    vc.Set(GCTithi.CalcTithiDate(va.gyear, va.masa, va.paksa, va.tithi % 15, earth))

#===========================================================================
#
#===========================================================================

def Gregorian2Gaurabda(vc, va, earth):
    day = GCDayData()
    day.DayCalc(vc, earth)
    va.masa = day.MasaCalc(vc, earth)
    va.tithi = day.nTithi
    va.gyear = day.nGaurabdaYear

def CalcEndDate(m_earth, vcStart, vaStart, vcEnd, vaEnd, nType, nCount):
    if nType==1:
        vcEnd.Set(vcStart)
        if nCount > 30240: nCount = 30240
        vcEnd.AddDays(nCount)
        Gregorian2Gaurabda(vcEnd, vaEnd, m_earth)
    elif nType==2:
        vcEnd.Set(vcStart)
        if nCount > 4320: nCount = 4320
        vcEnd.AddDays(nCount*7)
        Gregorian2Gaurabda(vcEnd, vaEnd, m_earth)
    elif nType==3:
        vcEnd.Set(vcStart)
        if nCount > 1080: nCount = 1080
        vcEnd.month += nCount
        while vcEnd.month > 12:
            vcEnd.year += 1
            vcEnd.month -= 12
        Gregorian2Gaurabda(vcEnd, vaEnd, m_earth)
    elif nType == 4:
        vcEnd.Set(vcStart)
        if nCount > 90: nCount = 90
        vcEnd.year += nCount
        Gregorian2Gaurabda(vcEnd, vaEnd, m_earth)
    elif nType == 5:
        vaEnd.Set(vaStart)
        if nCount > 30240: nCount = 30240
        vaEnd.tithi += nCount
        while vaEnd.tithi >= 30:
            vaEnd.tithi-=30
            vaEnd.masa+=1
        while vaEnd.masa >= 12:
            vaEnd.masa -= 12
            vaEnd.gyear+=1
        Gaurabda2Gregorian(vaEnd, vcEnd, m_earth)
    elif nType==6:
        vaEnd.Set(vaStart)
        if nCount > 1080: nCount = 1080
        vaEnd.masa = MasaToComboMasa(vaEnd.masa)
        if vaEnd.masa == MasaId.ADHIKA_MASA:
            vcEnd.Set(vcStart)
            vcEnd.month += nCount
            while vcEnd.month > 12:
                vcEnd.year+=1
                vcEnd.month -= 12
            Gregorian2Gaurabda(vcEnd, vaEnd, m_earth)
            vaEnd.tithi = vaStart.tithi
            Gaurabda2Gregorian(vaEnd, vcEnd, m_earth)
        else:
            vaEnd.masa += nCount
            while vaEnd.masa >= 12:
                vaEnd.masa -= 12
                vaEnd.gyear+=1
            vaEnd.masa = GCCalendar.ComboMasaToMasa(vaEnd.masa)
            Gaurabda2Gregorian(vaEnd, vcEnd, m_earth)
    elif nType==7:
        vaEnd.Set(vaStart)
        if nCount > 90: nCount = 90
        vaEnd.gyear += nCount
        Gaurabda2Gregorian(vaEnd, vcEnd, m_earth)
    return 1

def ComboMasaToMasa(nComboMasa):
    return 12 if nComboMasa == 12 else (nComboMasa + 11) % 12

def MasaToComboMasa(nMasa):
    return 12 if nMasa == 12 else (nMasa + 1) % 12

def writeFirstDayXml(xml,loc,vc):

    vcStart = GCGregorianDate(date=GetFirstDayOfYear(loc.GetEarthData(), vcStart.year))
    vcStart.InitWeekDay()

    # write
    xml.write("<xml>\n")
    xml.write("\t<request name=\"FirstDay\" version=\"")
    xml.write(GCStrings.getString(130))
    xml.write("\">\n")
    xml.write("\t\t<arg name=\"longitude\" val=\"")
    xml.write(str(loc.m_fLongitude))
    xml.write("\" />\n")
    xml.write("\t\t<arg name=\"latitude\" val=\"")
    xml.write(str(loc.m_fLatitude))
    xml.write("\" />\n")
    xml.write("\t\t<arg name=\"year\" val=\"")
    xml.write(str(vcStart.year))
    xml.write("\" />\n")
    xml.write("\t</request>\n")
    xml.write("\t<result name=\"FirstDay_of_GaurabdaYear\">\n")
    xml.write("\t\t<firstday date=\"")
    xml.write(str(vcStart))
    xml.write("\" dayweekid = \"")
    xml.write(str(vcStart.dayOfWeek))
    xml.write("\" dayweek=\"")
    xml.write(GCStrings.getString(vcStart.dayOfWeek))
    xml.write("\" />\n")
    xml.write("\t</result>\n")
    xml.write("</xml>\n")
    return 0


def unittests():
    GCUT.info('calendar')
    e = EARTHDATA()
    e.longitude_deg = 27.0
    e.latitude_deg = 45.0
    e.tzone = 1.0
    vc = Today()
    a = GetGaurabdaYear(vc,e)
    GCUT.msg('Gaurabda year:' + str(a))

    va = GCGaurabdaDate()
    Gregorian2Gaurabda(vc,va,e)
    GCUT.msg('Vatime: ' + str(va))

    va.prevMasa()
    va.tithi = 0
    vc2 = GCGregorianDate()
    Gaurabda2Gregorian(va,vc2,e)
    GCUT.msg('vctime2:' + str(vc2) + ' --> ' + str(va))



from .TCalendar import TCalendar
from .TAppDay import TAppDay
from .TCoreEvents import TCoreEvents
from .TMasaList import TMasaList
from .TToday import TToday
from .GCGregorianDate import GCGregorianDate,Today
from .GCLocation import GCLocation
from .GCLocationList import FindLocation,GetLocationsForCountry
from .GCCountry import GetCountries
from .GCTimeZone import GetTimeZones



import io
from . import GCUT as GCUT
from . import GCStrings as GCStrings
from . import GCDisplaySettings as GCDisplaySettings

SBTF_TEXT = 1
SBTF_RTF = 2
SBTF_HTML = 3
SBTF_XML = 4

class GCStringBuilder:
    def __init__(self,targetStream):
        self.fontSizeH1 = 28
        self.fontSizeH2 = 24
        self.fontSizeText = 20
        self.fontSizeNote = 18
        self.Format = SBTF_TEXT
        self.Target = targetStream

    def AppendString(self,text):
        self.Target.write(text)

    def AppendBoldString(self,text):
        if self.Format == SBTF_TEXT:
            self.Target.write(text)
        elif self.Format == SBTF_RTF:
            self.Target.write("{\\b ")
            self.Target.write(text)
            self.Target.write("}")
        elif self.Format == SBTF_HTML:
            self.Target.write("<b>")
            self.Target.write(text)
            self.Target.write("</b>")

    def AppendLine(self, text=''):
        if self.Format == SBTF_TEXT:
            self.Target.write(text)
            self.Target.write("\r\n")
        elif self.Format == SBTF_RTF:
            self.Target.write("\r\n\\par ")
            self.Target.write(text)
        elif self.Format == SBTF_HTML:
            self.Target.write("<br>")
            self.Target.write(text)

    def initWithString(self, str):
        self.Target.write(str)

    def AppendHeader1(self, text):
        if self.Format == SBTF_TEXT:
            lt = int((80 - len(text))/2)
            self.Target.write(''.rjust(lt,'=') + text + ''.rjust(lt,'='))
            self.Target.write('\r\n')
        elif self.Format == SBTF_RTF:
            self.Target.write("\\par\\pard{{\\f2\\fs{} {} }}".format(self.fontSizeH1,text))

    def AppendHeader3(self, text, fillChar='='):
        lt = int((80 - len(text))/2)
        t = ''.rjust(lt,fillChar) + text + ''.rjust(lt,fillChar)
        if self.Format == SBTF_TEXT:
            self.Target.write(t)
            self.Target.write('\r\n')
        elif self.Format == SBTF_RTF:
            self.Target.write("\\par\\pard{{\\f2\\fs{} {} }}".format(self.fontSizeH1,t))
        elif self.Format == SBTF_HTML:
            self.Target.write("<span style='color:#110033'>")
            self.Target.write(text)
            self.Target.write('</span>')


    def AppendHeader2(self, text):
        if self.Format == SBTF_TEXT:
            lt = int((80 - len(text))/2)
            self.Target.write(''.rjust(lt,'-') + text + ''.rjust(lt,'-'))
            self.Target.write('\r\n')
        elif self.Format == SBTF_RTF:
            self.Target.write("\\par\\pard{{\\f2\\fs{} {} }}".format(self.fontSizeH2,text))

    def AppendDocumentHeader(self):
        if self.Format == SBTF_TEXT:
            pass
        elif self.Format == SBTF_RTF:
            color_table = ''
            file_name = os.path.join(os.path.dirname(__file__), 'res', 'rtf_color_table.txt')
            with open(file_name) as rfile:
                color_table = rfile.read()
            self.Target.write("{\\rtf1\\ansi\\ansicpg1252\\deff2\\deflang1033{\\fonttbl{\\f0\\fswiss\\fcharset0 Lucida Console;} {\\f1\\fswiss\\fcharset0 Arial;}{\\f2\\froman\\fprq2\\fcharset0 Book Antiqua;}} {\\colortbl ;")
            self.Target.write(color_table)
            self.Target.write("}{\\*\\generator GCAL;}\\viewkind4\\uc1\\pard\\f0\\fs20 ")
            self.Target.write("\\par\\pard\\f2\\fs")
            self.Target.write(str(self.fontSizeText))
            self.Target.write("\r\n")

    def AppendDocumentTail(self):
        if self.Format == SBTF_TEXT:
            pass
        elif self.Format == SBTF_RTF:
            self.Target.write("}")
        elif self.Format == SBTF_HTML:
            self.Target.write("</body>")
            self.Target.write("</html>")

    def AppendNote(self):
        self.AppendLine()
        self.AppendLine()
        self.AppendSeparatorWithWidth(80)
        self.AppendLine()
        self.AppendBoldString("Notes:")
        self.AppendLine()
        self.AppendLine()
        self.AppendLine("DST - Time is in \'Daylight Saving Time\'")
        self.AppendLine("LT  - Time is in \'Local Time\'\r\n")

        if GCDisplaySettings.getValue(9) > 0 or GCDisplaySettings.getValue(10) > 0 or GCDisplaySettings.getValue(11) > 0 or GCDisplaySettings.getValue(12) > 0:
            self.AppendLine("(*) - value at the moment of sunrise")

        # last line
        self.AppendLine()
        self.AppendString("Generated by ")
        self.AppendString(GCStrings.getString(132))

    def AppendSeparatorWithWidth(self,width):
        self.AppendString('-' * width);

    def AppendTwoColumnText(self,pText,pText2):
        if self.Format == SBTF_TEXT:
            str = '{}{}'.format(pText.ljust(17,' '), pText2).rstrip() + "\r\n"
        elif self.Format == SBTF_RTF:
            str = '\\par {}\\tab {}'.format(pText, pText2)
        self.Target.write(str)

def unittests():
    GCUT.info('string builder')
    s = io.StringIO()
    sb = GCStringBuilder(s)
    GCUT.msg('testing TEXT')
    sb.AppendString('aaa')
    sb.AppendDocumentHeader()
    sb.AppendDocumentTail()
    sb.AppendNote()
    sb.AppendSeparatorWithWidth(30)
    GCUT.msg('testing RTF')
    sb.AppendString('aaa')
    sb.AppendBoldString('aaa')
    sb.AppendDocumentHeader()
    sb.AppendDocumentTail()
    sb.AppendNote()
    sb.AppendSeparatorWithWidth(30)




class GCHorizontalCoords:
    def __init__(self):
    	self.azimut = 0.0
    	self.elevation = 0.0


class GCEquatorialCoords:
    def __init__(self):
    	self.rightAscension = 0.0
    	self.declination = 0.0

class GCEclipticalCoords:
    def __init__(self):
    	self.latitude = 0.0
    	self.longitude = 0.0
    	self.distance = 0.0



import flask
import logging
from flask import jsonify, request, Response
import io


from .GCLocationList import FindLocation, FindLocations,GetLocationsForCountry
from .GCCountry import GetCountries
from .GCGregorianDate import GCGregorianDate,Today
from .TCalendar import TCalendar

app = flask.Flask(__name__)


@app.route('/countries', methods=['GET'])
def getListCountries():
    return jsonify(GetCountries())

@app.route('/find-location', methods=['GET','POST'])
def findLocation():
    if request.method == 'GET':
        # get name of location
        name = request.args.get('name')
        country = request.args.get('country')
        e, s, c = FindLocations(name, country)
        obj = {
            'EQUALS': [a.data() for a in e],
            'STARTS': [a.data() for a in s],
            'CONTAINS': [a.data() for a in c]
        }
        return jsonify(obj)
    elif request.method == 'POST':
        content = request.get_json()
        name = None
        country = None
        if 'name' in content:
            name = content['name']
        if 'country' in content:
            country = content['country']
        e, s, c = FindLocations(name, country)
        obj = {
            'EQUALS': [a.data() for a in e],
            'STARTS': [a.data() for a in s],
            'CONTAINS': [a.data() for a in c]
        }
        return jsonify(obj)



@app.route('/calendar', methods=['GET','POST'])
def getCalendar():
    loca = {}
    date = {}
    period = None
    fmt = None
    req_data = None
    if request.method == 'GET':
        req_data = request.args
    elif request.method == 'POST':
        req_data = request.json
    else:
        return flask.make_response('Unknown method', 500)

    loca['city'] = req_data.get('city')
    loca['country'] = req_data.get('country')
    loca['latitude'] = req_data.get('latitude')
    loca['longitude'] = req_data.get('longitude')
    loca['tzname'] = req_data.get('tzname')
    date['year'] = req_data.get('year')
    date['month'] = req_data.get('month')
    date['day'] = req_data.get('day')
    period = req_data.get('period')
    fmt = req_data.get('format')
    
    if loca['latitude'] is None and loca['longitude'] is not None \
       or loca['latitude'] is not None and loca['longitude'] is None:
       return flask.make_response('Either both LATITUDE,LONGITUDE are valid or none of them', 500)

    if loca['city'] is None:
        return flask.make_response('city: - Name of location must be specified.', 500)

    if loca['country'] is None:
        return flask.make_response('country: Name of country must be specified.', 500)

    if loca['latitude'] is None:
        sp = FindLocation(city=loca['city'], country=loca['country'])
        if sp is None:
            return flask.make_response('Location with name \'{}\', country \'{}\' is not found in database.'.format(loca['name'], loca['country']), 500)
        loca['latitude'] = sp.m_fLatitude
        loca['longitude'] = sp.m_fLongitude
        loca['tzname'] = sp.m_strTimeZone
        loca['location'] = sp
    else:
        loca['latitude'] = float(loca['latitude'])
        loca['longitude'] = float(loca['longitude'])

    if loca['tzname'] is None:
        return flask.make_response('tz: Name of timezone must be specified.', 500)
    
    if date['year'] is None:
        d = Today()
        date['year'] = d.year
    else:
        date['year'] = int(date['year'])

    if date['month'] is None:
        date['month'] = 1
    else:
        date['month'] = int(date['month'])

    if date['day'] is None:
        date['day'] = 1
    else:
        date['day'] = int(date['day'])

    if period is None:
        return flask.make_response('p: Time period must be specified.', 500)
    try:
        period = int(period)
    except:
        return flask.make_response('p: Time period is number of days (integer number).', 500)
    if period<1:
        return flask.make_response('p: Time period must be greater than 0 days.', 500)
    if period>3653:
        return flask.make_response('p: Time period must be lower than 3654 days.', 500)

    tc = TCalendar()
    date2 = GCGregorianDate(year=date['year'], month=date['month'], day=date['day'])
    location = loca.get('location')
    if location is None:
        location = GCLocation(data={
            'city': loca['city'],
            'country': loca['country'],
            'latitude': loca['latitude'],
            'longitude': loca['longitude'],
            'tzname': loca['tzname']
        })
    tc.CalculateCalendar(location,date2,period)

    wf = io.StringIO()

    # save results in various formats
    if fmt == 'txt' or fmt=='text' or fmt=='plain':
        tc.write(wf, format='plain')
        return Response(wf.getvalue(), mimetype='text/plain')
    elif fmt=='html':
        tc.write(wf)
        return Response(wf.getvalue(), mimetype='text/html')
    elif fmt=='html-table':
        tc.write(wf, layout='table')
        return Response(wf.getvalue(), mimetype='text/html')
    elif fmt=='xml':
        tc.write(wf, format='xml')
        return Response(wf.getvalue(), mimetype='text/xml')
    else:
        return jsonify(tc.get_json_object())


def run_server(port=8047):
    app.run(port=port)

if __name__=='__main__':
    run_server()


from .GCGregorianDate import GCGregorianDate,Today
from .GCLocation import GCLocation
from . import GCMath as GCMath
from . import GCAyanamsha as GCAyanamsha
from .GCTime import GCTime
from .GCSunData import SUNDATA
from . import GCSunData as GCSunData
from . import GCStrings as GCStrings
from math import floor
from . import GCUT as GCUT

sankrantiDetermineType = 2

def GetSankrantiType():
    return sankrantiDetermineType

def SetSankrantiType(i):
    prev = sankrantiDetermineType
    sankrantiDetermineType = i
    return prev

def GetSankMethodName(i):
    snam = ["midnight to midnight",
        "sunrise to sunrise",
        "noon to noon",
        "sunset to sunset"]
    return snam[i]


#********************************************************************/
#  Finds next time when rasi is changed                             */
#                                                                   */
#  startDate - starting date and time, timezone member must be valid */
#  zodiac [out] - found zodiac sign into which is changed           */
#                                                                   */
#********************************************************************/
def GetNextSankranti(startDate):
    zodiac = 0
    step = 1.0
    count = 0
    prevday = GCGregorianDate()

    d = GCGregorianDate(date = startDate)

    prev = GCMath.putIn360( GCSunData.GetSunLongitude(d) - GCAyanamsha.GetAyanamsa(d.GetJulian()))
    prev_rasi = int(floor(prev / 30.0))

    while count < 20:
        prevday.Set(d)
        d.shour += step
        d.NormalizeHours()

        ld = GCMath.putIn360(GCSunData.GetSunLongitude(d) - GCAyanamsha.GetAyanamsa(d.GetJulian()))
        new_rasi = int(floor(ld/30.0))

        if prev_rasi != new_rasi:
            zodiac = new_rasi
            step *= 0.5
            d.Set(prevday)
            count+=1
            continue

    return d,zodiac


def writeXml(xml, loc, vcStart, vcEnd):
    dt = GCTime()
    zodiac = 0

    d = GCGregorianDate(date = vcStart)

    xml.write("<xml>\n")
    xml.write("\t<request name=\"Sankranti\" version=\"")
    xml.write(GCStrings.getString(130))
    xml.write("\">\n")
    xml.write("\t\t<arg name=\"longitude\" val=\"")
    xml.write(str(loc.m_fLongitude))
    xml.write("\" />\n")
    xml.write("\t\t<arg name=\"latitude\" val=\"")
    xml.write(str(loc.m_fLatitude))
    xml.write("\" />\n")
    xml.write("\t\t<arg name=\"timezone\" val=\"")
    xml.write(str(loc.m_fTimezone))
    xml.write("\" />\n")
    xml.write("\t\t<arg name=\"location\" val=\"")
    xml.write(loc.m_strName)
    xml.write("\" />\n")
    xml.write("\t\t<arg name=\"startdate\" val=\"")
    xml.write(repr(vcStart))
    xml.write("\" />\n")
    xml.write("\t\t<arg name=\"enddate\" val=\"")
    xml.write(repr(vcEnd))
    xml.write("\" />\n")
    xml.write("\t</request>\n")
    xml.write("\t<result name=\"SankrantiList\">\n")

    while d.IsBeforeThis(vcEnd):
        nextDate,zodiac = GetNextSankranti(d)
        d.Set(nextDate)
        d.InitWeekDay()
        xml.write("\t\t<sank date=\"")
        xml.write(str(d))
        xml.write("\" ")
        xml.write("dayweekid=\"")
        xml.write(str(d.dayOfWeek))
        xml.write("\" dayweek=\"")
        xml.write(GCStrings.getString(d.dayOfWeek))
        xml.write("\" ")

        dt.SetDegTime( 360 * d.shour )

        xml.write(" time=\"")
        xml.write(repr(dt))
        xml.write("\" >\n")
        xml.write("\t\t\t<zodiac sans=\"")
        xml.write(GCStrings.GetSankrantiName(zodiac))
        xml.write("\" eng=\"")
        xml.write(GCStrings.GetSankrantiNameEn(zodiac))
        xml.write("\" id=\"")
        xml.write(str(zodiac))
        xml.write("\" />\n")
        xml.write("\t\t</sank>\n")

        d.NextDay()
        d.NextDay()

    xml.write("\t</result>\n")
    xml.write("</xml>")

    return 1


def unittests():
    GCUT.info('sankranti')
    vc = Today()
    vc2 = GCGregorianDate()
    vc3 = GCGregorianDate(date = vc)
    vc3.AddDays(100)
    n = GetSankMethodName(GetSankrantiType())
    GCUT.msg('Sankranti Type: {}'.format(n))
    import io
    s = io.StringIO()
    clr = GCLocation()
    writeXml(s, clr, vc, vc3)
    GCUT.msg(s.getvalue())



from . import GCMath as GCMath
from . import GCTime as GCTime
from . import GCUT as GCUT
from . import GCAyanamsha as GCAyanamsha
from .GCEnums import FastType,MahadvadasiType
from . import GCPancangaDate as GCPancangaDate
from . import GCDisplaySettings as GCDisplaySettings
from . import GCStrings as GCStrings
from . import GCCountry as GCCountry
from . import GCGregorianDate as GCGregorianDate
from . import GCTimeZone as GCTimeZone
from . import GCEarthData as GCEarthData
from . import GCStringBuilder as GCStringBuilder
from . import GCMoonData as GCMoonData
from . import GCSunData as GCSunData
from . import GCConjunction as GCConjunction
from . import GCNaksatra as GCNaksatra
from . import GCSankranti as GCSankranti
from . import GCDayData as GCDayData
from . import GCTithi as GCTithi
from . import GCYoga as GCYoga
from . import GCLocationList as GCLocationList
from . import GCEventList as GCEventList
from . import GCCalendar as GCCalendar
import TMasaList
import TCalendar
import TAppDay
import TCoreEvents
import TToday
import os

os.mkdir('test')

GCMath.unittests()
GCTime.unittests()
GCUT.info('enums')
GCUT.val(FastType.FAST_NULL,0, 'FastType')
GCUT.val(MahadvadasiType.EV_VIJAYA,0x110, 'MahadvadasiType')
GCAyanamsha.unittests()
GCPancangaDate.unittests()
GCDisplaySettings.unittests()
GCStrings.unittests()
GCCountry.unittests()
GCGregorianDate.unittests()
GCTimeZone.unittests()
GCEarthData.unittests()
GCStringBuilder.unittests()
GCMoonData.unittests()
GCSunData.unittests()
GCConjunction.unittests()
GCNaksatra.unittests()
GCSankranti.unittests()
GCTithi.unittests()
GCDayData.unittests()
GCYoga.unittests()
GCLocationList.unittests()
GCEventList.unittests()
GCCalendar.unittests()
# retype to if True: ... in case you need unittests for specific engine
if True: TMasaList.unittests()
if True: TCalendar.unittests()
if True: TAppDay.unittests()
if True: TCoreEvents.unittests()
TToday.unittests()



import math

from . import GCEarthData as GCEarthData
from . import GCGregorianDate as GCGregorianDate
from . import GCTime as GCTime
from . import GCCoords as GCCoords
from . import GCUT as GCUT
from . import GCMath as GCMath
from . import GCAyanamsha as GCAyanamsha

class MOONDATA:
    def __init__(self):
        #   latitude from nodal (draconic) phase
        self.latitude_deg = 0.0
        #   longitude from sidereal motion
        self.longitude_deg = 0.0
        self.radius = 0.0
        self.rektaszension = 0.0
        self.declination = 0.0
        self.parallax = 0.0
        self.elevation = 0.0
        self.azimuth = 0.0

    def CalculateEcliptical(self,jdate):
        arg_lr = [
             [ 0, 0, 1, 0], [ 2, 0,-1, 0],     [ 2, 0, 0, 0],     [ 0, 0, 2, 0],
             [ 0, 1, 0, 0],     [ 0, 0, 0, 2],     [ 2, 0,-2, 0],     [ 2,-1,-1, 0],
             [ 2, 0, 1, 0],     [ 2,-1, 0, 0],     [ 0, 1,-1, 0],     [ 1, 0, 0, 0],
             [ 0, 1, 1, 0],     [ 2, 0, 0,-2],     [ 0, 0, 1, 2],     [ 0, 0, 1,-2],
             [ 4, 0,-1, 0],     [ 0, 0, 3, 0],     [ 4, 0,-2, 0],     [ 2, 1,-1, 0],
             [ 2, 1, 0, 0],     [ 1, 0,-1, 0],     [ 1, 1, 0, 0],     [ 2,-1, 1, 0],
             [ 2, 0, 2, 0],     [ 4, 0, 0, 0],     [ 2, 0,-3, 0],     [ 0, 1,-2, 0],
             [ 2, 0,-1, 2],     [ 2,-1,-2, 0],     [ 1, 0, 1, 0],     [ 2,-2, 0, 0],
             [ 0, 1, 2, 0],     [ 0, 2, 0, 0],     [ 2,-2,-1, 0],     [ 2, 0, 1,-2],
             [ 2, 0, 0, 2],     [ 4,-1,-1, 0],     [ 0, 0, 2, 2],     [ 3, 0,-1, 0],
             [ 2, 1, 1, 0],     [ 4,-1,-2, 0],     [ 0, 2,-1, 0],     [ 2, 2,-1, 0],
             [ 2, 1,-2, 0],     [ 2,-1, 0,-2],     [ 4, 0, 1, 0],     [ 0, 0, 4, 0],
             [ 4,-1, 0, 0],     [ 1, 0,-2, 0],     [ 2, 1, 0,-2],     [ 0, 0, 2,-2],
             [ 1, 1, 1, 0],     [ 3, 0,-2, 0],     [ 4, 0,-3, 0],     [ 2,-1, 2, 0],
             [ 0, 2, 1, 0],     [ 1, 1,-1, 0],     [ 2, 0, 3, 0],     [ 2, 0,-1,-2]]

        arg_b = [[ 0, 0, 0, 1],     [ 0, 0, 1, 1],     [ 0, 0, 1,-1],     [ 2, 0, 0,-1],
                [ 2, 0,-1, 1],     [ 2, 0,-1,-1],     [ 2, 0, 0, 1],     [ 0, 0, 2, 1],
                [ 2, 0, 1,-1],     [ 0, 0, 2,-1],  [ 2,-1, 0,-1],     [ 2, 0,-2,-1],
                [ 2, 0, 1, 1],     [ 2, 1, 0,-1],     [ 2,-1,-1, 1],     [ 2,-1, 0, 1],
                [ 2,-1,-1,-1],     [ 0, 1,-1,-1],     [ 4, 0,-1,-1],     [ 0, 1, 0, 1],
                [ 0, 0, 0, 3],     [ 0, 1,-1, 1],     [ 1, 0, 0, 1],     [ 0, 1, 1, 1],
                [ 0, 1, 1,-1],     [ 0, 1, 0,-1],     [ 1, 0, 0,-1],     [ 0, 0, 3, 1],
                [ 4, 0, 0,-1],     [ 4, 0,-1, 1],     [ 0, 0, 1,-3],     [ 4, 0,-2, 1],
                [ 2, 0, 0,-3],     [ 2, 0, 2,-1],     [ 2,-1, 1,-1],     [ 2, 0,-2, 1],
                [ 0, 0, 3,-1],     [ 2, 0, 2, 1],     [ 2, 0,-3,-1],     [ 2, 1,-1, 1],
                [ 2, 1, 0, 1],     [ 4, 0, 0, 1],     [ 2,-1, 1, 1],     [ 2,-2, 0,-1],
                [ 0, 0, 1, 3],     [ 2, 1, 1,-1],     [ 1, 1, 0,-1],     [ 1, 1, 0, 1],
                [ 0, 1,-2,-1],     [ 2, 1,-1,-1],     [ 1, 0, 1, 1],     [ 2,-1,-2,-1],
                [ 0, 1, 2, 1],     [ 4, 0,-2,-1],     [ 4,-1,-1,-1],     [ 1, 0, 1,-1],
                [ 4, 0, 1,-1],     [ 1, 0,-1,-1],     [ 4,-1, 0,-1],     [ 2,-2, 0, 1]]
        sigma_r = [-20905355, -3699111, -2955968,  -569925,
               48888,    -3149,   246158,  -152138,
             -170733,  -204586,  -129620,   108743,
              104755,    10321,        0,    79661,
              -34782,   -23210,   -21636,    24208,
               30824,    -8379,   -16675,   -12831,
              -10445,   -11650,    14403,    -7003,
                   0,    10056,     6322,    -9884,
                5751,        0,    -4950,     4130,
                   0,    -3958,        0,     3258,
                2616,    -1897,    -2117,     2354,
                   0,        0,    -1423,    -1117,
               -1571,    -1739,        0,    -4421,
                   0,        0,        0,        0,
                1165,        0,        0,     8752]

        sigma_l = [6288774, 1274027,  658314,  213618,
            -185116, -114332,   58793,   57066,
              53322,   45758,  -40923,  -34720,
             -30383,   15327,  -12528,   10980,
              10675,   10034,    8548,   -7888,
              -6766,   -5163,    4987,    4036,
               3994,    3861,    3665,   -2689,
              -2602,    2390,   -2348,    2236,
              -2120,   -2069,    2048,   -1773,
              -1595,    1215,   -1110,    -892,
               -810,     759,    -713,    -700,
                691,     596,     549,     537,
                520,    -487,    -399,    -381,
                351,    -340,     330,     327,
               -323,     299,     294,       0]

        sigma_b = [5128122,  280602,  277693,  173237,
              55413,   46271,   32573,   17198,
               9266,    8822,    8216,    4324,
               4200,   -3359,    2463,    2211,
               2065,   -1870,    1828,   -1794,
              -1749,   -1565,   -1491,   -1475,
              -1410,   -1344,   -1335,    1107,
               1021,     833,     777,     671,
                607,     596,     491,    -451,
                439,     422,     421,    -366,
               -351,     331,     315,     302,
               -283,    -229,     223,     223,
               -220,    -220,    -185,     181,
               -177,     176,     166,    -164,
                132,    -119,     115,     107]

        t = (jdate - 2451545.0)/36525.0

        #  (* mean elongation of the moon
        d = 297.8502042+(445267.1115168+(-0.0016300+(1.0/545868-1.0/113065000*t)*t)*t)*t

        #  (* mean anomaly of the sun
        m = 357.5291092+(35999.0502909+(-0.0001536+1.0/24490000*t)*t)*t

        #  (* mean anomaly of the moon
        ms = 134.9634114+(477198.8676313+(0.0089970+(1.0/69699-1.0/1471200*t)*t)*t)*t

        #  (* argument of the longitude of the moon
        f = 93.2720993+(483202.0175273+(-0.0034029+(-1.0/3526000+1.0/863310000*t)*t)*t)*t

        #  (* correction term due to excentricity of the earth orbit
        e = 1.0+(-0.002516-0.0000074*t)*t

        #  (* mean longitude of the moon
        ls = 218.3164591+(481267.88134236+(-0.0013268+(1.0/538841-1.0/65194000*t)*t)*t)*t

        #  (* arguments of correction terms
        a1 = 119.75+131.849*t
        a2 = 53.09+479264.290*t
        a3 = 313.45+481266.484*t

        sr = 0
        for i in range(60):
            temp =sigma_r[i]*GCMath.cosDeg( arg_lr[i][0]*d +arg_lr[i][1]*m +arg_lr[i][2]*ms +arg_lr[i][3]*f)
            if math.fabs(arg_lr[i][1])==1: temp = temp*e
            if math.fabs(arg_lr[i][1])==2: temp = temp*e*e
            sr =sr+temp

        sl = 0
        for i in range(60):
            temp =sigma_l[i]*GCMath.sinDeg( arg_lr[i][0]*d +arg_lr[i][1]*m +arg_lr[i][2]*ms +arg_lr[i][3]*f)
            if math.fabs(arg_lr[i][1])==1: temp =temp*e
            if math.fabs(arg_lr[i][1])==2: temp =temp*e*e
            sl =sl+temp

        #  (* correction terms
        sl =sl +3958*GCMath.sinDeg(a1) \
            +1962*GCMath.sinDeg(ls-f) \
            +318*GCMath.sinDeg(a2)
        sb = 0
        for i in range(60):
            temp = sigma_b[i]*GCMath.sinDeg( arg_b[i][0]*d + arg_b[i][1]*m + arg_b[i][2]*ms + arg_b[i][3]*f)
            if math.fabs(arg_b[i][1])==1: temp=temp*e
            if math.fabs(arg_b[i][1])==2: temp=temp*e*e
            sb=sb+temp

        #  (* correction terms
        sb =sb - 2235*GCMath.sinDeg(ls) + 382*GCMath.sinDeg(a3)  + 175*GCMath.sinDeg(a1-f) + 175*GCMath.sinDeg(a1+f)  +127*GCMath.sinDeg(ls-ms) - 115*GCMath.sinDeg(ls+ms)

        coords = GCCoords.GCEclipticalCoords()

        coords.longitude = ls+sl/1000000
        coords.latitude = sb/1000000
        coords.distance = 385000.56+sr/1000

        return coords

    def Calculate(self,jdate,earth):
        crd = self.CalculateEcliptical(jdate)
        eqc,crd = GCEarthData.eclipticalToEquatorialCoords(crd, jdate)

        self.radius = crd.distance
        self.longitude_deg =crd.longitude
        self.latitude_deg = crd.latitude

        self.rektaszension = eqc.rightAscension
        self.declination = eqc.declination

    def calc_horizontal(self,date, longitude, latitude):
        h = GCMath.putIn360(GCEarthData.star_time(date) - self.rektaszension+longitude)

        self.azimuth = GCMath.rad2deg( math.atan2(GCMath.sinDeg(h), GCMath.cosDeg(h)*GCMath.sinDeg(latitude) - GCMath.tanDeg(self.declination)*GCMath.cosDeg(latitude) ))

        self.elevation = GCMath.rad2deg( math.asin( GCMath.sinDeg(latitude) * GCMath.sinDeg(self.declination) + GCMath.cosDeg(latitude) * GCMath.cosDeg(self.declination) * GCMath.cosDeg(h)))

    def  correct_position(self, jdate, latitude, longitude, height):
        b_a=0.99664719

        u =GCMath.arcTanDeg(b_a*b_a*GCMath.tanDeg(latitude))
        rho_sin =b_a*GCMath.sinDeg(u)+height/6378140.0*GCMath.sinDeg(latitude)
        rho_cos =GCMath.cosDeg(u)+height/6378140.0*GCMath.cosDeg(latitude)

        self.parallax = GCMath.arcSinDeg(GCMath.sinDeg(8.794/3600)/(MoonDistance(jdate) / GCMath.AU))

        h = GCEarthData.star_time(jdate) - longitude - self.rektaszension;
        delta_alpha = GCMath.arcTanDeg( (-rho_cos*GCMath.sinDeg(self.parallax)*GCMath.sinDeg(h)) / (GCMath.cosDeg(self.declination) - rho_cos*GCMath.sinDeg(self.parallax)*GCMath.cosDeg(h)))
        self.rektaszension = self.rektaszension+delta_alpha
        self.declination =GCMath.arcTanDeg( (( GCMath.sinDeg(self.declination) -rho_sin*GCMath.sinDeg(self.parallax))*GCMath.cosDeg(delta_alpha))/ ( GCMath.cosDeg(self.declination) -rho_cos*GCMath.sinDeg(self.parallax)*GCMath.cosDeg(h)))

    def getTopocentricEquatorial(self, obs, jdate):
        b_a = 0.99664719
        tec = GCCoords.GCEquatorialCoords()

        altitude = 0

        #   geocentric position of observer on the earth surface
        #   10.1 - 10.3
        u = GCMath.arcTanDeg(b_a * b_a * GCMath.tanDeg(obs.latitude_deg))
        rho_sin = b_a * GCMath.sinDeg(u) + altitude / 6378140.0 * GCMath.sinDeg(obs.latitude_deg)
        rho_cos = GCMath.cosDeg(u) + altitude / 6378140.0 * GCMath.cosDeg(obs.latitude_deg)

        #   equatorial horizontal paralax
        #   39.1
        parallax = GCMath.arcSinDeg(GCMath.sinDeg(8.794 / 3600) / (self.radius / GCMath.AU))

        #   geocentric hour angle of the body
        h = GCEarthData.star_time(jdate) + obs.longitude_deg - self.rektaszension


        #   39.2
        delta_alpha = GCMath.arcTanDeg( (-rho_cos * GCMath.sinDeg(self.parallax) * GCMath.sinDeg(h)) / (GCMath.cosDeg(self.declination) - rho_cos * GCMath.sinDeg(self.parallax) * GCMath.cosDeg(h)))
        tec.rightAscension = self.rektaszension + delta_alpha
        tec.declination = GCMath.arcTanDeg( ((GCMath.sinDeg(self.declination) - rho_sin * GCMath.sinDeg(self.parallax)) * GCMath.cosDeg(delta_alpha)) / (GCMath.cosDeg(self.declination) - rho_cos * GCMath.sinDeg(self.parallax) * GCMath.cosDeg(h)))

        return tec



def MoonCalcElevation(e,vc):
    moon = MOONDATA()
    d = vc.GetJulianComplete()
    moon.Calculate(d, e)
    moon.correct_position(d, e.latitude_deg, e.longitude_deg, 0)
    moon.calc_horizontal(d, e.longitude_deg, e.latitude_deg)

    return moon.elevation

def CalcMoonTimes(e,vc,nDaylightSavingShift):
    rise = GCTime.GCTime()
    set = GCTime.GCTime()

    UT = 0.0
    i = 0
    prev_elev = 0.0
    nType = 0
    nFound = 0

    rise.SetValue(-1)
    set.SetValue(-1)

    #   inicializacia prvej hodnoty ELEVATION
    vc.shour = (-nDaylightSavingShift - 1.0) / 24.0
    prev_elev = MoonCalcElevation(e, vc)

    #   prechod cez vsetky hodiny
    UT = - 0.1 - nDaylightSavingShift
    while UT <= (24.1 - nDaylightSavingShift):
        vc.shour = UT / 24.0
        elev = MoonCalcElevation(e, vc)
        if prev_elev * elev <= 0.0:
            if prev_elev <= 0.0:
                nType = 0x1
            else:
                nType = 0x2

            a = UT - 1.0
            ae = prev_elev
            b = UT
            be = elev
            for i in range(20):
                c = (a + b) / 2.0
                vc.shour = c / 24.0
                ce = MoonCalcElevation(e, vc)
                if ae*ce <= 0.0:
                    be = ce
                    b = c
                else:
                    ae = ce
                    a = c

            if nType == 1:
                rise.SetDayTime((c + nDaylightSavingShift)/24.0)
            else:
                set.SetDayTime((c + nDaylightSavingShift)/24.0)
            nFound |= nType
            if nFound == 0x3: break
        prev_elev = elev
        UT+=1.0
    return rise,set

def GetNextMoonRasi(ed,startDate):
    nextDate = GCGregorianDate.GCGregorianDate(date=startDate)
    phi = 30.0
    jday = startDate.GetJulianComplete()
    moon = MOONDATA()
    d = GCGregorianDate.GCGregorianDate(date=startDate)
    ayanamsa = GCAyanamsha.GetAyanamsa(jday)
    scan_step = 0.5
    prev_naks = 0
    new_naks = -1

    xj = 0.0
    xd = GCGregorianDate.GCGregorianDate()

    moon.Calculate(jday, ed)
    l1 = GCMath.putIn360(moon.longitude_deg - ayanamsa)
    prev_naks = int(GCMath.Floor(l1 / phi))

    counter = 0
    while counter < 20:
        xj = jday
        xd.Set(d)

        jday += scan_step
        d.shour += scan_step
        if d.shour > 1.0:
            d.shour -= 1.0
            d.NextDay()

        moon.Calculate(jday, ed)
        l2 = GCMath.putIn360(moon.longitude_deg - ayanamsa);
        new_naks = int(GCMath.Floor(l2/phi))
        if prev_naks != new_naks:
            jday = xj
            d.Set(xd)
            scan_step *= 0.5
            counter+=1
            continue
        else:
            l1 = l2
    nextDate.Set(d)
    return new_naks,nextDate


def GetNextRise(e, vc, bRise):
    nFound = 0
    h = [0, 0, 0]
    hour = 1/24.0
    startHour = vc.shour

    track = GCGregorianDate.GCGregorianDate(date = vc)
    track.NormalizeValues()

    #   inicializacia prvej hodnoty ELEVATION
    h[0] = MoonCalcElevation(e, track)
    track.shour += hour
    h[1] = MoonCalcElevation(e, track)
    track.shour += hour
    h[2] = MoonCalcElevation(e, track)

    for c in range(24):
        has_change = False
        if bRise:
            has_change = h[1] < 0.0 and h[2] > 0.0
        else:
            has_change = h[1] > 0.0 and h[2] < 0.0
        if has_change:
            a = (h[2] - h[1]) / hour
            b = h[2] - a * track.shour
            track.shour = - b / a
            track.NormalizeValues()
            return track

        h[0] = h[1]
        h[1] = h[2]
        track.shour += hour
        h[2] = MoonCalcElevation(e, track)

    return track



def MoonDistance(jdate):
    arg_lr = [
         [ 0, 0, 1, 0], [ 2, 0,-1, 0], [ 2, 0, 0, 0], [ 0, 0, 2, 0], [ 0, 1, 0, 0],
         [ 0, 0, 0, 2], [ 2, 0,-2, 0], [ 2,-1,-1, 0], [ 2, 0, 1, 0], [ 2,-1, 0, 0],
         [ 0, 1,-1, 0], [ 1, 0, 0, 0], [ 0, 1, 1, 0], [ 2, 0, 0,-2], [ 0, 0, 1, 2],
         [ 0, 0, 1,-2], [ 4, 0,-1, 0], [ 0, 0, 3, 0], [ 4, 0,-2, 0], [ 2, 1,-1, 0],
         [ 2, 1, 0, 0], [ 1, 0,-1, 0], [ 1, 1, 0, 0], [ 2,-1, 1, 0], [ 2, 0, 2, 0],
         [ 4, 0, 0, 0], [ 2, 0,-3, 0], [ 0, 1,-2, 0], [ 2, 0,-1, 2], [ 2,-1,-2, 0],
         [ 1, 0, 1, 0], [ 2,-2, 0, 0], [ 0, 1, 2, 0], [ 0, 2, 0, 0], [ 2,-2,-1, 0],
         [ 2, 0, 1,-2], [ 2, 0, 0, 2], [ 4,-1,-1, 0], [ 0, 0, 2, 2], [ 3, 0,-1, 0],
         [ 2, 1, 1, 0], [ 4,-1,-2, 0], [ 0, 2,-1, 0], [ 2, 2,-1, 0], [ 2, 1,-2, 0],
         [ 2,-1, 0,-2], [ 4, 0, 1, 0], [ 0, 0, 4, 0], [ 4,-1, 0, 0], [ 1, 0,-2, 0],
         [ 2, 1, 0,-2], [ 0, 0, 2,-2], [ 1, 1, 1, 0], [ 3, 0,-2, 0], [ 4, 0,-3, 0],
         [ 2,-1, 2, 0], [ 0, 2, 1, 0], [ 1, 1,-1, 0], [ 2, 0, 3, 0], [ 2, 0,-1,-2]
       ]

    sigma_r = [
       -20905355, -3699111, -2955968,  -569925,    48888,    -3149,   246158,
         -152138,  -170733,  -204586,  -129620,   108743,   104755,    10321,
               0,    79661,   -34782,   -23210,   -21636,    24208,    30824,
           -8379,   -16675,   -12831,   -10445,   -11650,    14403,    -7003,
               0,    10056,     6322,    -9884,     5751,        0,    -4950,
            4130,        0,    -3958,        0,     3258,     2616,    -1897,
           -2117,     2354,        0,        0,    -1423,    -1117,    -1571,
           -1739,        0,    -4421,        0,        0,        0,        0,
            1165,        0,        0,     8752
      ]


    t = (jdate - 2451545.0)/36525.0

    #  (* mean elongation of the moon
    d = 297.8502042+(445267.1115168+(-0.0016300+(1.0/545868-1.0/113065000*t)*t)*t)*t

    #  (* mean anomaly of the sun
    m = 357.5291092+(35999.0502909+(-0.0001536+1.0/24490000*t)*t)*t

    #  (* mean anomaly of the moon
    ms = 134.9634114+(477198.8676313+(0.0089970+(1.0/69699-1.0/1471200*t)*t)*t)*t

    #  (* argument of the longitude of the moon
    f = 93.2720993+(483202.0175273+(-0.0034029+(-1.0/3526000+1.0/863310000*t)*t)*t)*t

    #  (* correction term due to excentricity of the earth orbit
    e = 1.0+(-0.002516-0.0000074*t)*t

    #  (* mean longitude of the moon
    ls = 218.3164591+(481267.88134236+(-0.0013268+(1.0/538841-1.0/65194000*t)*t)*t)*t

    sr = 0

    for i in range(60):
        temp =sigma_r[i]*GCMath.cosDeg( arg_lr[i][0]*d + arg_lr[i][1]*m + arg_lr[i][2]*ms + arg_lr[i][3]*f)
        if math.fabs(arg_lr[i][1])==1: temp = temp*e
        if math.fabs(arg_lr[i][1])==2: temp = temp*e*e
        sr =sr+temp


    return 385000.56+sr/1000


def unittests():
    GCUT.info('moon data')
    ed = GCEarthData.EARTHDATA()
    ed.longitude_deg = 20.0
    ed.latitude_deg = 45.0
    ed.tzone = 1.0
    vc = GCGregorianDate.GCGregorianDate(2000,1,1,0.5)
    GCUT.msg('Moon distance: ' + str(MoonDistance(vc.GetJulianComplete())))
    nd = GetNextRise(ed,vc,True)
    GCUT.msg('Next rise: ' + repr(nd))
    nd = GetNextRise(ed, vc, False)
    GCUT.msg('Next set: ' + repr(nd))
    a,b = GetNextMoonRasi(ed,vc)
    GCUT.msg('next moon rasi: {} {}'.format(a,b))
    a,b = CalcMoonTimes(ed,vc,1.0)
    GCUT.msg('moon times: {} {}'.format(repr(a),repr(b)))



from . import GCMath as GCMath
from . import GCUT as GCUT
import math

class GCTime:
    def __init__(self,hour=0,min=0,sec=0,milli=0):
        self.hour = hour
        self.min = min
        self.sec = sec
        self.milli = milli

    def __repr__(self):
        return '{:02d}:{:02d}:{:02d}.{:03d}'.format(self.hour, self.min, self.sec, self.milli%1000)

    # offset in seconds
    def short(self,offset=0):
        h,m,s = self.normalize(self.hour,self.min,self.sec + offset)
        return '{:02d}:{:02d}'.format(h,m)

    def normalize(self,h,m,s):
        while s>59:
            s-=60
            m+=1
        while s<0:
            s+=60
            m-=1
        while m>59:
            m-=60
            h+=1
        while m<0:
            m+=60
            h-=1
        return h,m,s

    def SetDayTime(self,d):
        time_hr = 0.0

        # hour
        time_hr = d * 24
        self.hour = int( math.floor(time_hr) )

        # minute
        time_hr -= self.hour
        time_hr *= 60
        self.min = int( math.floor(time_hr) )

        # second
        time_hr -= self.min
        time_hr *= 60
        self.sec = int( math.floor(time_hr) )

        # miliseconds
        time_hr -= self.sec
        time_hr *= 1000
        self.mili = int( math.floor(time_hr) )

    def SetValue(self,i):
        self.hour,self.min,self.sec = self.normalize(0,0,i)
        self.milli = i-math.floor(i)

    def Set(self,d,offset=0):
        self.hour = d.hour
        self.min = d.min
        self.sec = d.sec
        self.mili = d.mili
        if offset!=0:
            self.AddMinutes(offset)

    def IsGreaterThan(self,dt):
        return self.Seconds()>dt.Seconds()

    def IsLessThan(self, dt):
        return self.Seconds()<dt.Seconds()

    def IsGreaterOrEqualThan(self, dt):
        return self.Seconds()>=dt.Seconds()

    def IsLessOrEqualThan(self, dt):
        return self.Seconds()<=dt.Seconds()

    def Seconds(self):
        return self.hour*3600 + self.min*60 + self.sec + self.milli/1000

    def AddMinutes(self,mn):
        self.hour,self.min,self.sec = self.normalize(self.hour,self.min,self.sec + int(mn*60))

    def AddSeconds(self,sc):
        self.hour,self.min,self.sec = self.normalize(self.hour,self.min,self.sec + int(sc))

    def GetDayTime(self):
        return ((self.hour*60.0 + self.min)*60.0 + self.sec) / 86400.0

    #  Conversion time from DEGREE fromat to H:M:S:MS format
    #
    #  time - output
    #  time_deg - input time in range 0 deg to 360 deg
    #             where 0 deg = 0:00 AM and 360 deg = 12:00 PM
    #
    def SetDegTime(self, time_deg):
        time_hr = 0.0

        time_deg = GCMath.putIn360(time_deg)

        # hour
        time_hr = time_deg / 360 * 24
        self.hour = int( math.floor(time_hr) )

        # minute
        time_hr -= self.hour
        time_hr *= 60
        self.min = int( math.floor(time_hr) )

        # second
        time_hr -= self.min
        time_hr *= 60
        self.sec = int( math.floor(time_hr) )

        # miliseconds
        time_hr -= self.sec
        time_hr *= 1000
        self.mili = int( math.floor(time_hr) )

    def ToLongTimeString(self):
        return '{:02d}:{:02d}:{:02d}'.format(self.hour, self.min, self.sec)


def unittests():
    dt1 = GCTime(hour=1,min=23)
    dt2 = GCTime(hour=2,min=30)
    dt3 = GCTime(hour=5,min=10)

    GCUT.info('GCTime class')
    GCUT.val(dt1.ToLongTimeString(),'01:23:00','long time string')
    GCUT.val(dt3.IsGreaterThan(dt2),True,'greater than')

    dt2.AddMinutes(176)
    GCUT.val(dt2.ToLongTimeString(),'05:26:00', 'add minutes')
    GCUT.val(dt3.IsLessThan(dt2),True,'less than')



from .GCEnums.CoreEventType import *
from .GCEnums.KalaType import *
from .GCEnums.GCDS import *
from .GCGregorianDate import GCGregorianDate,Today
from .GCLocation import GCLocation
from . import GCTimeZone as GCTimeZone
from .GCSunData import SUNDATA,CalculateKala
from .GCMoonData import MOONDATA,GetNextMoonRasi
from .GCStringBuilder import GCStringBuilder,SBTF_RTF,SBTF_TEXT
from . import GCTithi as GCTithi
from . import GCNaksatra as GCNaksatra
from . import GCSankranti as GCSankranti
from . import GCConjunction as GCConjunction
from . import GCAyanamsha as GCAyanamsha
from . import GCYoga as GCYoga
from . import GCRasi as GCRasi
from . import GCStrings as GCStrings
from . import GCDisplaySettings as GCDisplaySettings
from . import GCLayoutData as GCLayoutData
from math import ceil,floor
from . import GCUT as GCUT
import datetime


class TDayEvent:
    def __init__(self):
        self.nType = 0
        self.nData = 0
        self.Time = GCGregorianDate()
        self.nDst = 0
        self.julianDay = 0

    def Set(self,de):
        nType = de.nType
        nData = de.nData
        Time.Set(de.Time)
        nDst = de.nDst

    def EventText(self):
        if self.nType == CCTYPE_S_ARUN:
            return 'arunodaya'
        elif self.nType == CCTYPE_S_RISE:
            return "sunrise"
        elif self.nType == CCTYPE_S_NOON:
            return "noon"
        elif self.nType == CCTYPE_S_SET:
            return "sunset"
        elif self.nType == CCTYPE_TITHI:
            return GCStrings.GetTithiName(self.nData) + " Tithi starts"
        elif self.nType == CCTYPE_NAKS:
            return GCStrings.GetNaksatraName(self.nData) + " Naksatra starts"
        elif self.nType == CCTYPE_YOGA:
            return GCStrings.GetYogaName(self.nData) + " Yoga starts"
        elif self.nType == CCTYPE_SANK:
            return "Sun enters {}".format(GCStrings.GetSankrantiName(self.nData))
        elif self.nType == CCTYPE_CONJ:
            return "conjunction in {} rasi".format(GCStrings.GetSankrantiName(self.nData))
        elif self.nType == CCTYPE_KALA_START:
            return GCStrings.GetKalaName(self.nData) + " starts"
        elif self.nType == CCTYPE_KALA_END:
            return GCStrings.GetKalaName(self.nData) + " ends"
        elif self.nType == CCTYPE_M_RISE:
            return "moonrise"
        elif self.nType == CCTYPE_M_SET:
            return "moonset"
        elif self.nType == CCTYPE_ASCENDENT:
            return GCStrings.GetSankrantiName(self.nData) + " ascendent"
        elif self.nType == CCTYPE_M_RASI:
            return "Moon enters {}".format(GCStrings.GetSankrantiName(self.nData))
        else:
            return ''

class TCoreEvents:
    header_text = {
        CCTYPE_DATE: " DATE ",
        CCTYPE_S_ARUN: " SUNRISE, SUNSET ",
        CCTYPE_S_RISE: " SUNRISE, SUNSET ",
        CCTYPE_S_NOON: " SUNRISE, SUNSET ",
        CCTYPE_S_SET: " SUNRISE, SUNSET ",
        CCTYPE_TITHI: " TITHI ",
        CCTYPE_NAKS: " NAKSATRA ",
        CCTYPE_SANK: " SANKRANTI ",
        CCTYPE_CONJ: " SUN-MOON CONJUNCTION ",
        CCTYPE_YOGA: " YOGA ",
        CCTYPE_KALA_START: " KALAS ",
        CCTYPE_KALA_END: " KALAS ",
        CCTYPE_M_RISE: " MOONRISE, MOONSET ",
        CCTYPE_M_SET: " MOONRISE, MOONSET ",
        CCTYPE_M_RASI: " MOON RASI ",
        CCTYPE_ASCENDENT: " ASCENDENT "
    }
    def __init__(self):
        self.m_vcStart = GCGregorianDate()
        self.m_vcEnd = GCGregorianDate()
        self.m_options = 0
        self.m_location = GCLocation()
        self.p_events = []
        self.b_sorted = True

    def AddEvent(self, inTime, inType, inData, inDst):
        p = TDayEvent()
        self.p_events.append(p)


        if inDst == 1:
            if (inTime.shour >= 2/24.0):
                inTime.shour += 1/24.0
                inTime.NormalizeValues()
                p.nDst = 1
        elif inDst == 2:
            inTime.shour += 1/24.0
            inTime.NormalizeValues()
            p.nDst = 1
        elif inDst == 3:
            if (inTime.shour <= 2/24.0):
                inTime.shour += 1/24.0
                inTime.NormalizeValues()
                p.nDst = 1
        p.Time.Set(inTime)
        p.Time.InitWeekDay()
        p.julianDay = p.Time.GetJulianComplete()
        p.nData = inData
        p.nType = inType
        return True

    def Sort(self):
        self.p_events = sorted(self.p_events, key = lambda k: k.julianDay)

    def CalculateEvents(self,loc,vcStart,vcEnd):
        sun = SUNDATA()
        ndst = 0

        self.m_location.Set(loc)
        self.m_vcStart.Set(vcStart)
        self.m_vcEnd.Set(vcEnd)

        vcNext = GCGregorianDate()
        earth = loc.GetEarthData()

        vc = GCGregorianDate(date = vcStart)
        vcAdd = GCGregorianDate(date = vcStart)
        vcAdd.InitWeekDay()

        previousLongitude = -100
        todayLongitude = 0
        fromTimeLimit = 0

        while vcAdd.IsBeforeThis(vcEnd):
            if (GCDisplaySettings.getValue(COREEVENTS_SUN)):
                ndst = GCTimeZone.determineDaylightChange(vcAdd, loc.m_nTimezoneId)
                sun.SunCalc(vcAdd, earth)

                vcAdd.shour = sun.arunodaya.GetDayTime()
                self.AddEvent(vcAdd, CCTYPE_S_ARUN, 0, ndst)

                vcAdd.shour = sunRise = sun.rise.GetDayTime()
                self.AddEvent(vcAdd, CCTYPE_S_RISE, 0, ndst)

                vcAdd.shour = sun.noon.GetDayTime()
                self.AddEvent(vcAdd, CCTYPE_S_NOON, 0, ndst)

                vcAdd.shour = sunSet = sun.set.GetDayTime()
                self.AddEvent(vcAdd, CCTYPE_S_SET, 0, ndst)
            else:
                ndst = GCTimeZone.determineDaylightChange(vcAdd, loc.m_nTimezoneId)
                sun.SunCalc(vcAdd, earth)
                sunRise = sun.rise.GetDayTime()
                sunSet = sun.set.GetDayTime()

            if (GCDisplaySettings.getValue(COREEVENTS_ASCENDENT)):
                todayLongitude = sun.longitude_deg
                vcAdd.shour = sunRise
                todaySunriseHour = sunRise
                if (previousLongitude < -10):
                    prevSunrise = GCGregorianDate(date = vcAdd)
                    prevSunrise.PreviousDay()
                    sun.SunCalc(prevSunrise, earth)
                    previousSunriseHour = sun.rise.GetDayTime() - 1
                    previousLongitude = sun.longitude_deg
                    fromTimeLimit = 0

                jd = vcAdd.GetJulianComplete()
                ayan = GCAyanamsha.GetAyanamsa(jd)
                r1 = GCMath.putIn360(previousLongitude - ayan) / 30
                r2 = GCMath.putIn360(todayLongitude - ayan) / 30

                while(r2 > r1 + 13):
                    r2 -= 12.0
                while(r2 < r1 + 11):
                    r2 += 12.0

                a = (r2 - r1) / (todaySunriseHour - previousSunriseHour)
                b = r2 - a * todaySunriseHour

                tr = ceil(r1)
                for tr in range(ceil(r1),ceil(r2)):
                    tm = ( tr - b ) / a
                    if (tm > fromTimeLimit):
                        vcNext.Set(vcAdd)
                        vcNext.shour = tm
                        vcNext.NormalizeValues()
                        self.AddEvent(vcNext, CCTYPE_ASCENDENT, tr, ndst)

                previousLongitude = todayLongitude
                previousSunriseHour = todaySunriseHour - 1
                fromTimeLimit = previousSunriseHour

            if (GCDisplaySettings.getValue(COREEVENTS_RAHUKALAM)):
                r1,r2 = CalculateKala(sunRise, sunSet, vcAdd.dayOfWeek,KT_RAHU_KALAM)
                vcAdd.shour = r1
                self.AddEvent(vcAdd, CCTYPE_KALA_START, KT_RAHU_KALAM, ndst)
                vcAdd.shour = r2
                self.AddEvent(vcAdd, CCTYPE_KALA_END, KT_RAHU_KALAM, ndst)

            if (GCDisplaySettings.getValue(COREEVENTS_YAMAGHANTI)):
                r1,r2 = CalculateKala(sunRise, sunSet, vcAdd.dayOfWeek, KT_YAMA_GHANTI)
                vcAdd.shour = r1
                self.AddEvent(vcAdd, CCTYPE_KALA_START, KT_YAMA_GHANTI, ndst)
                vcAdd.shour = r2
                self.AddEvent(vcAdd, CCTYPE_KALA_END, KT_YAMA_GHANTI, ndst)

            if (GCDisplaySettings.getValue(COREEVENTS_GULIKALAM)):
                r1,r2 = CalculateKala(sunRise, sunSet, vcAdd.dayOfWeek, KT_GULI_KALAM)
                vcAdd.shour = r1
                self.AddEvent(vcAdd, CCTYPE_KALA_START, KT_GULI_KALAM, ndst)
                vcAdd.shour = r2
                self.AddEvent(vcAdd, CCTYPE_KALA_END, KT_GULI_KALAM, ndst)

            if (GCDisplaySettings.getValue(COREEVENTS_ABHIJIT_MUHURTA)):
                r1,r2 = CalculateKala(sunRise, sunSet, vcAdd.dayOfWeek, KT_ABHIJIT)
                if (r1 > 0 and r2 > 0):
                    vcAdd.shour = r1
                    self.AddEvent(vcAdd, CCTYPE_KALA_START, KT_ABHIJIT, ndst)
                    vcAdd.shour = r2
                    self.AddEvent(vcAdd, CCTYPE_KALA_END, KT_ABHIJIT, ndst)

            vcAdd.NextDay()

        if (GCDisplaySettings.getValue(COREEVENTS_TITHI)):
            vcAdd.Set(vc)
            vcAdd.shour = 0.0
            while vcAdd.IsBeforeThis(vcEnd):
                nData = GCTithi.GetNextTithiStart(earth, vcAdd, vcNext)
                if (vcNext.GetDayInteger() < vcEnd.GetDayInteger()):
                    vcNext.InitWeekDay()
                    ndst = GCTimeZone.determineDaylightChange(vcNext, loc.m_nTimezoneId)
                    self.AddEvent(vcNext, CCTYPE_TITHI, nData, ndst)
                else:
                    break
                vcAdd.Set(vcNext)
                vcAdd.shour += 0.2
                if (vcAdd.shour >= 1.0):
                    vcAdd.shour -= 1.0
                    vcAdd.NextDay()

        if (GCDisplaySettings.getValue(COREEVENTS_NAKSATRA)):
            vcAdd.Set(vc)
            vcAdd.shour = 0.0
            while vcAdd.IsBeforeThis(vcEnd):
                nData = GCNaksatra.GetNextNaksatra(earth, vcAdd, vcNext)
                if (vcNext.GetDayInteger() < vcEnd.GetDayInteger()):
                    vcNext.InitWeekDay()
                    ndst = GCTimeZone.determineDaylightChange(vcNext, loc.m_nTimezoneId)
                    self.AddEvent(vcNext, CCTYPE_NAKS, nData, ndst)
                else:
                    break
                vcAdd.Set(vcNext)
                vcAdd.shour += 0.2
                if (vcAdd.shour >= 1.0):
                    vcAdd.shour -= 1.0
                    vcAdd.NextDay()

        if (GCDisplaySettings.getValue(COREEVENTS_YOGA)):
            vcAdd.Set(vc)
            vcAdd.shour = 0.0
            while vcAdd.IsBeforeThis(vcEnd):
                nData = GCYoga.GetNextYogaStart(earth, vcAdd, vcNext)
                if (vcNext.GetDayInteger() < vcEnd.GetDayInteger()):
                    vcNext.InitWeekDay()
                    ndst = GCTimeZone.determineDaylightChange(vcNext, loc.m_nTimezoneId)
                    self.AddEvent(vcNext, CCTYPE_YOGA, nData, ndst)
                else:
                    break
                vcAdd.Set(vcNext)
                vcAdd.shour += 0.2
                if (vcAdd.shour >= 1.0):
                    vcAdd.shour -= 1.0
                    vcAdd.NextDay()

        if (GCDisplaySettings.getValue(COREEVENTS_SANKRANTI)):
            vcAdd.Set(vc)
            vcAdd.shour = 0.0
            while vcAdd.IsBeforeThis(vcEnd):
                date,nData = GCSankranti.GetNextSankranti(vcAdd)
                vcNext.Set(date)
                if (vcNext.GetDayInteger() < vcEnd.GetDayInteger()):
                    vcNext.InitWeekDay()
                    ndst = GCTimeZone.determineDaylightChange(vcNext, loc.m_nTimezoneId)
                    self.AddEvent(vcNext, CCTYPE_SANK, nData, ndst)
                else:
                    break
                vcAdd.Set(vcNext)
                vcAdd.NextDay()

        if (GCDisplaySettings.getValue(COREEVENTS_MOONRASI)):
            vcAdd.Set(vc)
            vcAdd.shour = 0.0
            while vcAdd.IsBeforeThis(vcEnd):
                nData = GetNextMoonRasi(earth, vcAdd, vcNext)
                if (vcNext.GetDayInteger() < vcEnd.GetDayInteger()):
                    vcNext.InitWeekDay()
                    ndst = GCTimeZone.determineDaylightChange(vcNext, loc.m_nTimezoneId)
                    self.AddEvent(vcNext, CCTYPE_M_RASI, nData, ndst)
                else:
                    break
                vcAdd.Set(vcNext)
                vcAdd.shour += 0.5
                vcAdd.NormalizeValues()

        if (GCDisplaySettings.getValue(COREEVENTS_CONJUNCTION)):
            vcAdd.Set(vc)
            vcAdd.shour = 0.0
            while vcAdd.IsBeforeThis(vcEnd):
                dlong = GCConjunction.GetNextConjunctionEx(vcAdd, vcNext, True, earth)
                if (vcNext.GetDayInteger() < vcEnd.GetDayInteger()):
                    vcNext.InitWeekDay()
                    ndst = GCTimeZone.determineDaylightChange(vcNext, loc.m_nTimezoneId)
                    self.AddEvent(vcNext, CCTYPE_CONJ, GCRasi.GetRasi(dlong, GCAyanamsha.GetAyanamsa(vcNext.GetJulianComplete())), ndst)
                else:
                    break
                vcAdd.Set(vcNext)
                vcAdd.NextDay()

        if (GCDisplaySettings.getValue(COREEVENTS_MOON)):
            vcAdd.Set(vc)
            vcAdd.shour = 0.0
            while vcAdd.IsBeforeThis(vcEnd):
                vcNext.Set(MOONDATA.GetNextRise(earth, vcAdd, True))
                self.AddEvent(vcNext, CCTYPE_M_RISE, 0, ndst)

                vcNext.Set(MOONDATA.GetNextRise(earth, vcNext, False))
                self.AddEvent(vcNext, CCTYPE_M_SET, 0, ndst)

                vcNext.shour += 0.05
                vcNext.NormalizeValues()
                vcAdd.Set(vcNext)

        if self.b_sorted:
            self.Sort()

    def formatText(self,stream):
        sb = GCStringBuilder(stream)
        sb.Format = SBTF_TEXT

        stream.write("Events from {} to {}.\r\n\r\n".format( self.m_vcStart, self.m_vcEnd))
        stream.write("{}\r\n\r\n".format(self.m_location.m_strFullName))

        prevd = GCGregorianDate()
        prevd.day = 0
        prevd.month = 0
        prevd.year = 0
        prevt = -1

        last_header = ''
        for dnr in self.p_events:
            new_header = ''
            if self.b_sorted:
                new_header = " {} - {} ".format(dnr.Time, GCStrings.GetDayOfWeek(dnr.Time.dayOfWeek))
            else:
                new_header = header_text[dnr.nType]
            if last_header != new_header:
                sb.AppendLine()
                sb.AppendHeader3(new_header)
                sb.AppendLine()
                last_header = new_header

            stream.write("            {} {}    {}".format(dnr.Time.time_str(), GCStrings.GetDSTSignature(dnr.nDst), dnr.EventText()))
            sb.AppendLine()

        sb.AppendLine()
        sb.AppendNote()
        return 1

    def formatXml(self,strXml):
        strXml.write("<xml>\r\n<program version=\"{}\">\r\n<location longitude=\"{}\" latitude=\"{}\" timezone=\"{}\" dst=\"{}\" />\n".format(GCStrings.getString(130), self.m_location.m_fLongitude, self.m_location.m_fLatitude , self.m_location.m_fTimezone, GCTimeZone.GetTimeZoneName(self.m_location.m_nTimezoneId)))

        for dnr in self.p_events:
            strXml.write("  <event type=\"{}\" date=\"{}\" time=\"{}\" dst=\"{}\" />\n".format(dnr.EventText(), str(dnr.Time), dnr.Time.time_str(), dnr.nDst))

        strXml.write("</xml>\n")
        return 1

    def formatRtf(self,stream):
        sb = GCStringBuilder(stream)
        sb.Format = SBTF_RTF
        sb.fontSizeH1 = GCLayoutData.textSizeH1
        sb.fontSizeH2 = GCLayoutData.textSizeH2
        sb.fontSizeText = GCLayoutData.textSizeText
        sb.fontSizeNote = GCLayoutData.textSizeNote

        sb.AppendDocumentHeader()

        sb.AppendHeader1("Events")

        stream.write("\\par from {} to {}.\\par\r\n\\par\r\n".format( self.m_vcStart, self.m_vcEnd))
        stream.write("{}\\par\r\n\\par\r\n".format(self.m_location.m_strFullName))

        prevd = GCGregorianDate()
        prevd.day = 0
        prevd.month = 0
        prevd.year = 0
        prevt = -1

        last_header = ''
        for dnr in self.p_events:
            new_header = ''
            if self.b_sorted:
                new_header = " {} - {} ".format(dnr.Time, GCStrings.GetDayOfWeek(dnr.Time.dayOfWeek))
            else:
                new_header = header_text[dnr.nType]
            if last_header != new_header:
                sb.AppendLine()
                sb.AppendHeader2(new_header)
                sb.AppendLine()
                last_header = new_header

            stream.write("\\par            {} {}    {}".format(dnr.Time.time_str(), GCStrings.GetDSTSignature(dnr.nDst), dnr.EventText()))

        sb.AppendLine()
        sb.AppendNote()
        sb.AppendDocumentTail()
        return 1


    def writeHtml(self,stream):
        stream.write("<html>\n<head>\n<title>Core Events</title>\n\n")
        stream.write("<style>\n<!--\nbody {\n  font-family:Verdana;\n  font-size:11pt;\n}\n\ntd.hed {\n  font-size:11pt;\n  font-weight:bold;\n")
        stream.write("  background:#aaaaaa;\n  color:white;\n  text-align:center;\n  vertical-align:center;\n  padding-left:15pt;\n  padding-right:15pt;\n")
        stream.write("  padding-top:5pt;\n  padding-bottom:5pt;\n}\n-->\n</style>\n")
        stream.write("</head>\n")
        stream.write("<body>\n\n")
        stream.write("<h1 align=center>Events</h1>\n<p align=center>From {} to {}.</p>\n\n".format( self.m_vcStart, self.m_vcEnd))

        stream.write("<p align=center>{}</p>\n".format(self.m_location.m_strFullName))

        prevd = GCGregorianDate()
        prevd.day = 0
        prevd.month = 0
        prevd.year = 0
        prevt = -1

        stream.write("<table align=center><tr>\n")
        last_header = ''
        new_header = ''
        for dnr in self.p_events:
            if self.b_sorted:
                new_header = " {} - {} ".format(dnr.Time, GCStrings.GetDayOfWeek(dnr.Time.dayOfWeek))
            else:
                new_header = self.header_text[dnr.nType]
            if last_header != new_header:
                stream.write(f"<td class=\"hed\" colspan=2>{new_header}</td></tr>\n<tr>\n")
                last_header = new_header

            stream.write("<td>{}</td><td>{}</td></tr><tr>\n".format(dnr.EventText(), dnr.Time.time_str() ))

        stream.write("</tr></table>\n")
        stream.write("<hr align=center width=\"50%%\">\n<p align=center>Generated by {}</p>".format(GCStrings.getString(130)))
        stream.write("</body>\n</html>\n")
        return 1

    def write(self,stream,format='html',layout='list'):
        if format=='plain':
            self.formatText(stream)
        elif format=='rtf':
            self.formatRtf(stream)
        elif format=='xml':
            self.formatXml(stream)
        elif format=='html':
            self.writeHtml(stream)
            
def unittests():
    GCUT.info('core events results')
    loc = GCLocation(data={
        'latitude': 48.150002,
        'longitude': 17.116667,
        'tzid': 321,
        'name': 'Bratislava, Slovakia'
    })
    earth = loc.GetEarthData()
    today = Today()
    future = GCGregorianDate(date = today, addDays=100)

    tc = TCoreEvents()


    print('start calculate', datetime.datetime.now())
    tc.CalculateEvents(loc,today,future)
    print('end calculate', datetime.datetime.now())

    with open('test/events.xml','wt') as wf:
        tc.formatXml(wf)
    with open('test/events.txt','wt') as wf:
        tc.formatText(wf)
    with open('test/events.rtf','wt') as wf:
        tc.formatRtf(wf)
    with open('test/events.html','wt') as wf:
        tc.writeHtml(wf)



DISP_ALWAYS = -1
CAL_ARUN_TIME  = 1
CAL_ARUN_TITHI = 0
CAL_SUN_RISE = 2
CAL_SUN_NOON = 34
CAL_SUN_SET = 3
CAL_MOON_RISE = 4
CAL_MOON_SET = 5
CAL_KSAYA = 7
CAL_VRDDHI = 8
CAL_SUN_LONG = 9
CAL_MOON_LONG = 10
CAL_AYANAMSHA = 11
CAL_JULIAN = 12
CATURMASYA_PURNIMA = 13
CATURMASYA_PRATIPAT = 14
CATURMASYA_EKADASI = 15
CAL_SANKRANTI = 16
CAL_EKADASI_PARANA = 17
CAL_MASA_CHANGE = 21
CAL_FEST_0 = 22
CAL_FEST_1 = 23
CAL_FEST_2 = 24
CAL_FEST_3 = 25
CAL_FEST_4 = 26
CAL_FEST_5 = 27
CAL_FEST_6 = 28
CAL_DST_CHANGE = 35
GENERAL_FIRST_DOW = 40
COREEVENTS_SUN = 52
COREEVENTS_TITHI = 53
COREEVENTS_NAKSATRA = 54
COREEVENTS_SANKRANTI = 55
COREEVENTS_CONJUNCTION = 56
COREEVENTS_RAHUKALAM = 57
COREEVENTS_YAMAGHANTI = 58
COREEVENTS_GULIKALAM = 59
COREEVENTS_MOON = 60
COREEVENTS_MOONRASI = 61
COREEVENTS_ASCENDENT = 62
COREEVENTS_SORT = 63
COREEVENTS_ABHIJIT_MUHURTA = 64
COREEVENTS_YOGA = 65



from .GCMoonData import MOONDATA
from .GCSunData import SUNDATA, GetSunLongitude
from .GCGregorianDate import GCGregorianDate,Today
from . import GCEarthData as GCEarthData
from . import GCAyanamsha as GCAyanamsha
from . import GCMath as GCMath
from . import GCUT as GCUT
from math import floor

def GetNextYogaStart(ed, startDate, nextDate):
    phi = 40.0/3.0
    jday = startDate.GetJulianComplete()
    moon = MOONDATA()
    d = GCGregorianDate(date=startDate)
    xd = GCGregorianDate()
    scan_step = 0.5
    prev_tit = 0
    new_tit = -1
    ayanamsha = GCAyanamsha.GetAyanamsa(jday)
    moon.Calculate(jday, ed)
    sunl = GetSunLongitude(d)
    l1 = GCMath.putIn360( moon.longitude_deg + sunl - 2*ayanamsha)
    prev_tit = int(floor(l1/phi))

    counter = 0
    while counter < 20:
        xj = jday
        xd.Set(d)

        jday += scan_step
        d.shour += scan_step
        d.NormalizeHours()

        moon.Calculate(jday, ed)
        sunl = GetSunLongitude(d)
        l2 = GCMath.putIn360( moon.longitude_deg + sunl - 2*ayanamsha)
        new_tit = int(floor(l2/phi))

        if prev_tit != new_tit:
            jday = xj
            d.Set(xd)
            scan_step *= 0.5
            counter+=1
            continue
        else:
            l1 = l2
    nextDate.Set(d)

    return new_tit

def unittests():
    GCUT.info('yoga')
    e = GCEarthData.EARTHDATA()
    e.longitude_deg = 27.0
    e.latitude_deg = 45.0
    e.tzone = 1.0
    vc = Today()
    vc2 = GCGregorianDate()
    GetNextYogaStart(e,vc,vc2)
    print('Next yoga:', repr(vc2))



CMASYA_SYSTEM_PURNIMA = 13
CMASYA_SYSTEM_PRATIPAT = 14
CMASYA_SYSTEM_EKADASI = 15
CMASYA_SYSTEM_MASK = 0xff

CMASYA_MONTH_1 = 0x100
CMASYA_MONTH_2 = 0x200
CMASYA_MONTH_3 = 0x300
CMASYA_MONTH_4 = 0x400
CMASYA_MONTH_MASK = 0xf00

CMASYA_DAY_FIRST = 0x1000
CMASYA_DAY_LAST  = 0x2000
CMASYA_DAY_MASK  = 0xf000



from .GCGregorianDate import GCGregorianDate,Today
from .GCTime import GCTime
from . import GCLocation as GCLocation
from . import GCPancangaDate as GCPancangaDate
from . import GCEarthData as GCEarthData
from . import GCTithi as GCTithi
from .GCPancangaDate import GCGaurabdaDate
from .GCMoonData import MOONDATA
from .GCSunData import GetSunLongitude, SUNDATA
from .GCDayData import GCDayData, GetFirstDayOfYear, gGaurBeg
from . import GCMath as GCMath
from . import GCStrings as GCStrings
from math import floor
from . import GCUT as GCUT


def TITHI_EKADASI(a):
	return a in [10,25]

def TITHI_DVADASI(a):
	return a in [11,26]

def TITHI_TRAYODASI(a):
	return a in [12,27]

def TITHI_CATURDASI(a):
    return a in [13,28]

def TITHI_LESS_EKADASI(a):
    return a in [9,24,8,23]

def TITHI_LESS_DVADASI(a):
    return a in [10,25,9,24]

def TITHI_LESS_TRAYODASI(a):
    return a in [11,26,10,25]

def TITHI_FULLNEW_MOON(a):
    return a in [14,29]

def PREV_PREV_TITHI(a):
    return (a + 28) % 30

def PREV_TITHI(a):
    return (a + 29) % 30

def NEXT_TITHI(a):
    return (a + 1) % 30

def NEXT_NEXT_TITHI(a):
    return (a + 2) % 30

def TITHI_LESS_THAN(a,b):
    return a == PREV_TITHI(b) or a == PREV_PREV_TITHI(b)

def TITHI_GREAT_THAN(a,b):
    return a == NEXT_TITHI(b) or a == NEXT_NEXT_TITHI(b)

# TRUE when transit from A to B is between T and U
def TITHI_TRANSIT(t,u,a,b):
    return ((t==a) and (u==b)) or ((t==a) and (u==NEXT_TITHI(b))) or ((t==PREV_TITHI(a)) and (u == b))

#********************************************************************/
#                                                                   */
#   finds next time when starts next tithi                          */
#                                                                   */
#   timezone is not changed                                         */
#                                                                   */
#   return value: index of tithi 0..29                              */
#                 or -1 if failed                                   */
#********************************************************************/

def GetNextTithiStart(ed, startDate, nextDate, forward=True):
    phi = 12.0
    jday = startDate.GetJulianComplete()
    moon = MOONDATA()
    d = GCGregorianDate(date = startDate)
    xd = GCGregorianDate()
    dir = 1.0 if forward else -1.0
    scan_step = 0.5*dir
    prev_tit = 0
    new_tit = -1

    moon.Calculate(jday, ed)
    sunl = GetSunLongitude(d)
    l1 = GCMath.putIn360(moon.longitude_deg - sunl - 180.0)
    prev_tit = int(floor(l1/phi))

    counter = 0
    while counter < 20:
        xj = jday
        xd.Set(d)

        jday += scan_step
        d.shour += scan_step
        d.NormalizeHours()

        moon.Calculate(jday, ed)
        sunl = GetSunLongitude(d)
        l2 = GCMath.putIn360(moon.longitude_deg - sunl - 180.0)
        new_tit = int(floor(l2/phi))

        if prev_tit != new_tit:
            jday = xj
            d.Set(xd)
            scan_step *= 0.5
            counter+=1
            continue
        else:
            l1 = l2
    nextDate.Set(d)
    return new_tit


#********************************************************************/
#                                                                   */
#   finds previous time when starts next tithi                      */
#                                                                   */
#   timezone is not changed                                         */
#                                                                   */
#   return value: index of tithi 0..29                              */
#                 or -1 if failed                                   */
#********************************************************************/

def GetPrevTithiStart(ed, startDate, nextDate):
    return GetNextTithiStart(ed,startDate,nextDate,forward=False)


def GetTithiTimes(ed, vc, sunRise):
    d1 = GCGregorianDate()
    d2 = GCGregorianDate()

    vc.shour = sunRise
    GetPrevTithiStart(ed, vc, d1)
    GetNextTithiStart(ed, vc, d2)

    titBeg = d1.shour + d1.GetJulian() - vc.GetJulian()
    titEnd = d2.shour + d2.GetJulian() - vc.GetJulian()

    return titEnd - titBeg,titBeg,titEnd


#********************************************************************/
# Finds starting and ending time for given tithi                    */
#                                                                   */
# tithi is specified by Gaurabda year, masa, paksa and tithi number */
#      nGYear - 0..9999                                             */
#       nMasa - 0..12, 0-Madhusudana, 1-Trivikrama, 2-Vamana        */
#                      3-Sridhara, 4-Hrsikesa, 5-Padmanabha         */
#                      6-Damodara, 7-Kesava, 8-narayana, 9-Madhava  */
#                      10-Govinda, 11-Visnu, 12-PurusottamaAdhika   */
#       nPaksa -       0-Krsna, 1-Gaura                             */
#       nTithi - 0..14                                              */
#       earth  - used timezone                                      */
#                                                                   */
#********************************************************************/

def CalcTithiEnd(nGYear, nMasa, nPaksa, nTithi, earth, endTithi):
    d = GCGregorianDate(date = GetFirstDayOfYear(earth, nGYear + 1486))
    d.shour = 0.5
    d.tzone = earth.tzone

    return CalcTithiEndEx(d, nGYear, nMasa, nPaksa, nTithi, earth, endTithi)


def CalcTithiEndEx(vcStart, GYear, nMasa, nPaksa, nTithi, earth, endTithi):
    d = GCGregorianDate()
    dtemp = GCGregorianDate()
    day = GCDayData()
    tithi = 0
    counter = 0
    sunrise = 0.0
    start = GCGregorianDate()
    end = GCGregorianDate()
    start.shour = -1.0
    end.shour = -1.0
    start.day = start.month = start.year = -1
    end.day = end.month = end.year = -1

    d.Set(vcStart)

    i = 0
    while True:
        d.AddDays(13)
        day.DayCalc(d, earth)
        day.nMasa = day.MasaCalc(d, earth)
        gy = day.nGaurabdaYear
        i+=1
        if ((day.nPaksa == nPaksa) and (day.nMasa == nMasa)) or i > 30:
            break

    if i > 30:
        d.year = d.month = d.day = -1
        return d

    # we found masa and paksa
    # now we have to find tithi
    tithi = nTithi + nPaksa*15

    if day.nTithi == tithi:
        # loc1
        # find tithi juncts in this day and according to that times,
        # look in previous or next day for end and start of this tithi
        nType = 1
    else:
        if day.nTithi < tithi:
            # do increment of date until nTithi == tithi
            #   but if nTithi > tithi
            #       then do decrement of date
            counter = 0
            while counter < 30:
                d.NextDay()
                day.DayCalc(d, earth)
                if day.nTithi == tithi:
                    break
                if (day.nTithi < tithi ) and (day.nPaksa != nPaksa):
                    d.PreviousDay()
                    break
                if day.nTithi > tithi:
                    d.PreviousDay()
                    break
                counter+=1
            # somewhere is error
            if counter>=30:
                d.year = d.month = d.day = 0
                nType = 0
        else:
            # do decrement of date until nTithi <= tithi
            counter = 0
            while counter < 30:
                d.PreviousDay()
                day.DayCalc(d, earth)
                if day.nTithi == tithi:
                    break
                if (day.nTithi > tithi ) and (day.nPaksa != nPaksa):
                    break
                if day.nTithi < tithi:
                    break
                counter+=1
            # somewhere is error
            if counter >= 30:
                d.year = d.month = d.day = 0
                nType = 0

        if day.nTithi == tithi:
            # do the same as in loc1
            nType = 1
        else:
            # nTithi != tithi and nTithi < tithi
            # but on next day is nTithi > tithi
            # that means we will find start and the end of tithi
            # in this very day or on next day before sunrise
            nType = 2

    # now we know the type of day-accurancy
    # nType = 0 means, that we dont found a day
    # nType = 1 means, we find day, when tithi was present at sunrise
    # nType = 2 means, we found day, when tithi started after sunrise
    #                  but ended before next sunrise
    #
    sunrise = day.sun.sunrise_deg / 360 + earth.tzone/24

    if nType == 1:
        d1 = GCGregorianDate()
        d2 = GCGregorianDate()
        d.shour = sunrise
        GCTithi.GetPrevTithiStart(earth, d, d1)
        #d = d1
        #d.shour += 0.02
        GCTithi.GetNextTithiStart(earth, d, d2)

        endTithi.Set(d2)
        return d1
    elif nType == 2:
        d1 = GCGregorianDate()
        d2 = GCGregorianDate()
        d.shour = sunrise
        GCTithi.GetNextTithiStart(earth, d, d1)
        d.Set(d1)
        d.shour += 0.1
        d.NormalizeValues()
        GCTithi.GetNextTithiStart(earth, d, d2)

        endTithi.Set(d2)
        return d1

    # if nType == 0, then this algoritmus has some failure
    if nType == 0:
        d.year = 0
        d.month = 0
        d.day = 0
        d.shour = 0.0
        endTithi.Set(d)
    else:
        d.Set(start)
        endTithi.Set(end)

    return d

#********************************************************************/
#  Calculates Date of given Tithi                                   */
#********************************************************************/

def CalcTithiDate(nGYear, nMasa, nPaksa, nTithi, earth):
    i = 0
    gy = 0
    d = GCGregorianDate()
    dtemp = GCGregorianDate()
    day = GCDayData()
    tithi = 0
    counter = 0
    tmp = 0

    if nGYear >= 464 and nGYear < 572:
        tmp = gGaurBeg[(nGYear-464)*26 + nMasa*2 + nPaksa]
        d.month = (tmp & 0x3e0) >> 5
        d.day   = (tmp & 0x1f)
        d.year  = (tmp & 0xfffc00) >> 10
        d.tzone = earth.tzone
        d.NextDay()

        day.DayCalc(d, earth)
        day.nMasa = day.MasaCalc(d, earth)
        gy = day.nGaurabdaYear
    else:
        #d = GetFirstDayOfYear(earth, nGYear + 1486)
        d.day = 15
        d.month = 2 + nMasa
        d.year = nGYear + 1486
        if d.month > 12:
            d.month -= 12
            d.year+=1
        d.shour = 0.5
        d.tzone = earth.tzone

        i = 0
        while True:
            d.AddDays(13)
            day.DayCalc(d, earth)
            day.nMasa = day.MasaCalc(d, earth)
            gy = day.nGaurabdaYear
            i+=1
            if i>=30: break
            if (day.nPaksa == nPaksa) and (day.nMasa == nMasa): break

    if i >= 30:
        d.year = d.month = d.day = -1
        return d

    # we found masa and paksa
    # now we have to find tithi
    tithi = nTithi + nPaksa*15

    if day.nTithi == tithi:
        # loc1
        # find tithi juncts in this day and according to that times,
        # look in previous or next day for end and start of this tithi
        d.PreviousDay()
        day.DayCalc(d, earth)
        if (day.nTithi > tithi ) and (day.nPaksa != nPaksa):
            d.NextDay()
        return d

    if day.nTithi < tithi:
        # do increment of date until nTithi == tithi
        #   but if nTithi > tithi
        #       then do decrement of date
        counter = 0
        while counter < 16:
            d.NextDay()
            day.DayCalc(d, earth)
            if day.nTithi == tithi:
                return d
            if (day.nTithi < tithi ) and (day.nPaksa != nPaksa):
                return d
            if day.nTithi > tithi:
                return d
            counter+=1
        # somewhere is error
        d.year = d.month = d.day = 0
        return d
    else:
        # do decrement of date until nTithi <= tithi
        counter = 0
        while counter < 16:
            d.PreviousDay()
            day.DayCalc(d, earth)
            if day.nTithi == tithi:
                return d
            if (day.nTithi > tithi ) and (day.nPaksa != nPaksa):
                d.NextDay()
                return d
            if day.nTithi < tithi:
                d.NextDay()
                return d
            counter+=1
        # somewhere is error
        d.year = d.month = d.day = 0
        return d

    # now we know the type of day-accurancy
    # nType = 0 means, that we dont found a day
    # nType = 1 means, we find day, when tithi was present at sunrise
    # nType = 2 means, we found day, when tithi started after sunrise
    #                  but ended before next sunrise
    #

    return d

def writeXml(xml,loc,vc):
    date = GCGregorianDate()
    xml.write("<xml>\n")
    xml.write("\t<request name=\"Tithi\" version=\"")
    xml.write(GCStrings.getString(130))
    xml.write("\">\n")
    xml.write("\t\t<arg name=\"longitude\" val=\"")
    xml.write(str(loc.m_fLongitude))
    xml.write("\" />\n")
    xml.write("\t\t<arg name=\"latitude\" val=\"")
    xml.write(str(loc.m_fLatitude))
    xml.write("\" />\n")
    xml.write("\t\t<arg name=\"timezone\" val=\"")
    xml.write(str(loc.m_fTimezone))
    xml.write("\" />\n")
    xml.write("\t\t<arg name=\"startdate\" val=\"")
    xml.write(str(vc))
    xml.write("\" />\n")
    xml.write("\t</request>\n")
    xml.write("\t<result name=\"Tithi\">\n")

    d1 = GCGregorianDate()
    d2 = GCGregorianDate()
    dn = GCGregorianDate()
    dt = GCTime()
    earth = loc.GetEarthData()

    day = GCDayData()
    day.DayCalc(vc, earth)

    d = GCGregorianDate(date = vc)
    d.tzone = loc.m_fTimezone
    d.shour = day.sun.sunrise_deg/360.0 + loc.m_fTimezone/24.0

    GCTithi.GetPrevTithiStart(earth, d, d1)
    GCTithi.GetNextTithiStart(earth, d, d2)

    dt.SetDegTime( d1.shour * 360 )
    # start tithi at t[0]
    xml.write("\t\t<tithi\n\t\t\tid=\"")
    xml.write(str(day.nTithi))
    xml.write("\"\n")
    xml.write("\t\t\tname=\"")
    xml.write(GCStrings.GetTithiName(day.nTithi))
    xml.write("\"\n")
    xml.write("\t\t\tstartdate=\"")
    xml.write(str(d1))
    xml.write("\"\n")
    xml.write("\t\t\tstarttime=\"")
    xml.write(repr(dt))
    xml.write("\"\n")

    dt.SetDegTime( d2.shour * 360 )
    xml.write("\t\t\tenddate=\"")
    xml.write(str(d2))
    xml.write("\"\n")
    xml.write("\t\t\tendtime=\"")
    xml.write(repr(dt))
    xml.write("\"\n />")

    xml.write("\t</result>\n")
    xml.write("</xml>\n")

    return 1



def writeGaurabdaTithiXml(xml,loc,vaStart,vaEnd):
    gyearA = vaStart.gyear
    gyearB = vaEnd.gyear
    gmasa = vaStart.masa
    gpaksa = 1 if vaStart.tithi >= 15 else 0
    gtithi = vaStart.tithi % 15

    if gyearB < gyearA:
        gyearB = gyearA

    xml.write("<xml>\n")
    xml.write("\t<request name=\"Tithi\" version=\"")
    xml.write(GCStrings.getString(130))
    xml.write("\">\n")
    xml.write("\t\t<arg name=\"longitude\" val=\"")
    xml.write(str(loc.m_fLongitude))
    xml.write("\" />\n")
    xml.write("\t\t<arg name=\"latitude\" val=\"")
    xml.write(str(loc.m_fLatitude))
    xml.write("\" />\n")
    xml.write("\t\t<arg name=\"timezone\" val=\"")
    xml.write(str(loc.m_fTimezone))
    xml.write("\" />\n")
    if gyearA > 1500:
        xml.write("\t\t<arg name=\"year-start\" val=\"")
        xml.write(str(gyearA))
        xml.write("\" />\n")
        xml.write("\t\t<arg name=\"year-end\" val=\"")
        xml.write(str(gyearB))
        xml.write("\" />\n")
    else:
        xml.write("\t\t<arg name=\"gaurabdayear-start\" val=\"")
        xml.write(str(gyearA))
        xml.write("\" />\n")
        xml.write("\t\t<arg name=\"gaurabdayear-end\" val=\"")
        xml.write(str(gyearB))
        xml.write("\" />\n")
    xml.write("\t\t<arg name=\"masa\" val=\"")
    xml.write(str(gmasa))
    xml.write("\" />\n")
    xml.write("\t\t<arg name=\"paksa\" val=\"")
    xml.write(str(gpaksa))
    xml.write("\" />\n")
    xml.write("\t\t<arg name=\"tithi\" val=\"")
    xml.write(str(gtithi))
    xml.write("\" />\n")
    xml.write("\t</request>\n")
    xml.write("\t<result name=\"Tithi\">\n")


    earth = loc.GetEarthData()
    vcs = GCGregorianDate()
    vce = GCGregorianDate()
    today = GCGregorianDate()
    sun = SUNDATA()
    day = GCDayData()
    A = B = 0

    if gyearA > 1500:
        A = gyearA - 1487
        B = gyearB - 1485
    else:
        A = gyearA
        B = gyearB

    for A in range(A,B+1):
        vcs.Set(GCTithi.CalcTithiEnd(A, gmasa, gpaksa, gtithi, earth, vce))
        if gyearA > 1500:
            if (vcs.year < gyearA) or (vcs.year > gyearB):
                continue
        oTithi = gpaksa*15 + gtithi
        oMasa = gmasa
        oPaksa = gpaksa
        oYear = 0
        xml.write("\t<celebration\n")
        xml.write("\t\trtithi=\"")
        xml.write(GCStrings.GetTithiName(oTithi))
        xml.write("\"\n")
        xml.write("\t\trmasa=\"")
        xml.write(GCStrings.GetMasaName(oMasa))
        xml.write("\"\n")
        xml.write("\t\trpaksa=\"")
        xml.write('Gaura' if oPaksa else "Krsna")
        xml.write("\"\n")
        # test ci je ksaya
        today.Set(vcs)
        today.shour = 0.5
        sun.SunCalc(today, earth)
        sunrise = (sun.sunrise_deg + loc.m_fTimezone*15.0)/360
        if sunrise < vcs.shour:
            today.Set(vce)
            sun.SunCalc(today, earth)
            sunrise = (sun.sunrise_deg + loc.m_fTimezone*15.0)/360
            if sunrise < vce.shour:
                # normal type
                vcs.NextDay()
                xml.write("\t\ttype=\"normal\"\n")
            else:
                # ksaya
                vcs.NextDay()
                day.DayCalc(vcs, earth)
                oTithi = day.nTithi
                oPaksa = day.nPaksa
                oMasa = day.MasaCalc(vcs, earth)
                oYear = day.nGaurabdaYear
                xml.write("\t\ttype=\"ksaya\"\n")
        else:
            # normal, alebo prvy den vriddhi
            today.Set(vce)
            sun.SunCalc(today, earth)
            if (sun.sunrise_deg + loc.m_fTimezone*15.0)/360 < vce.shour:
                # first day of vriddhi type
                xml.write("\t\ttype=\"vriddhi\"\n")
            else:
                # normal
                xml.write("\t\ttype=\"normal\"\n")
        xml.write("\t\tdate=\"")
        xml.write(str(vcs))
        xml.write("\"\n")
        xml.write("\t\totithi=\"")
        xml.write(GCStrings.GetTithiName(oTithi))
        xml.write("\"\n")
        xml.write("\t\tomasa=\"")
        xml.write(GCStrings.GetMasaName(oMasa))
        xml.write("\"\n")
        xml.write("\t\topaksa=\"")
        xml.write('Gaura' if oPaksa else 'Krsna')
        xml.write("\"\n")
        xml.write("\t/>\n")

    xml.write("\t</result>\n")
    xml.write("</xml>\n")
    return 1

def writeGaurabdaNextTithiXml(xml,loc,vcStart,vaStart):
    gmasa = vaStart.masa
    gpaksa = 1 if vaStart.tithi >= 15 else 0
    gtithi = vaStart.tithi % 15

    xml.write("<xml>\n")
    xml.write("\t<request name=\"Tithi\" version=\"")
    xml.write(GCStrings.getString(130))
    xml.write("\">\n")
    xml.write("\t\t<arg name=\"longitude\" val=\"")
    xml.write(str(loc.m_fLongitude))
    xml.write("\" />\n")
    xml.write("\t\t<arg name=\"latitude\" val=\"")
    xml.write(str(loc.m_fLatitude))
    xml.write("\" />\n")
    xml.write("\t\t<arg name=\"timezone\" val=\"")
    xml.write(str(loc.m_fTimezone))
    xml.write("\" />\n")
    xml.write("\t\t<arg name=\"start date\" val=\"")
    xml.write(str(vcStart))
    xml.write("\" />\n")
    xml.write("\t\t<arg name=\"masa\" val=\"")
    xml.write(str(gmasa))
    xml.write("\" />\n")
    xml.write("\t\t<arg name=\"paksa\" val=\"")
    xml.write(str(gpaksa))
    xml.write("\" />\n")
    xml.write("\t\t<arg name=\"tithi\" val=\"")
    xml.write(str(gtithi))
    xml.write("\" />\n")
    xml.write("\t</request>\n")
    xml.write("\t<result name=\"Tithi\">\n")

    earth = loc.GetEarthData()
    vcs = GCGregorianDate()
    vce = GCGregorianDate()
    sun = SUNDATA()
    A = 0
    day = GCDayData()

    today = GCGregorianDate(date = vcStart)
    today.PreviousDay()
    vcStart.SubtractDays(15)
    for A in range(0,4):
        vcs.Set(GCTithi.CalcTithiEndEx(vcStart, 0, gmasa, gpaksa, gtithi, earth, vce))
        if not vcs.IsBeforeThis(today):
            oTithi = gpaksa*15 + gtithi
            oMasa = gmasa
            oPaksa = gpaksa
            oYear = 0
            xml.write("\t<celebration\n")
            xml.write("\t\trtithi=\"")
            xml.write(GCStrings.GetTithiName(oTithi))
            xml.write("\"\n")
            xml.write("\t\trmasa=\"")
            xml.write(GCStrings.GetMasaName(oMasa))
            xml.write("\"\n")
            xml.write("\t\trpaksa=\"")
            xml.write("Gaura" if oPaksa else "Krsna")
            xml.write("\"\n")
            # test ci je ksaya
            today.Set(vcs)
            today.shour = 0.5
            sun.SunCalc(today, earth)
            sunrise = (sun.sunrise_deg + loc.m_fTimezone*15.0)/360
            if sunrise < vcs.shour:
                today.Set(vce)
                sun.SunCalc(today, earth)
                sunrise = (sun.sunrise_deg + loc.m_fTimezone*15.0)/360
                if sunrise < vce.shour:
                    # normal type
                    vcs.NextDay()
                    xml.write("\t\ttype=\"normal\"\n")
                else:
                    # ksaya
                    vcs.NextDay()
                    day.DayCalc(vcs, earth)
                    oTithi = day.nTithi
                    oPaksa = day.nPaksa
                    oMasa = day.MasaCalc(vcs, earth)
                    oYear = day.nGaurabdaYear
                    xml.write("\t\ttype=\"ksaya\"\n")
            else:
                # normal, alebo prvy den vriddhi
                today.Set(vce)
                sun.SunCalc(today, earth)
                if (sun.sunrise_deg + loc.m_fTimezone*15.0)/360 < vce.shour:
                    # first day of vriddhi type
                    xml.write("\t\ttype=\"vriddhi\"\n")
                else:
                    # normal
                    xml.write("\t\ttype=\"normal\"\n")
            xml.write("\t\tdate=\"")
            xml.write(str(vcs))
            xml.write("\"\n")
            xml.write("\t\totithi=\"")
            xml.write(GCStrings.GetTithiName(oTithi))
            xml.write("\"\n")
            xml.write("\t\tomasa=\"")
            xml.write(GCStrings.GetMasaName(oMasa))
            xml.write("\"\n")
            xml.write("\t\topaksa=\"")
            xml.write("Gaura" if oPaksa else "Krsna")
            xml.write("\"\n")
            xml.write("\t/>\n")
            break
        else:
            vcStart.Set(vcs)
            vcs.NextDay()

    xml.write("\t</result>\n")
    xml.write("</xml>\n")
    return 1

# Finds first day of given masa and gaurabda year
def GetFirstDayOfMasa(earth,GYear,nMasa):
    return CalcTithiDate(GYear, nMasa, 0, 0, earth)


def unittests():
    GCUT.info('gc tithi')
    e = GCEarthData.EARTHDATA()
    e.longitude_deg = 27.0
    e.latitude_deg = 45.0
    e.tzone = 1.0
    vc = Today()
    vc2 = GCGregorianDate()
    vc3 = GCGregorianDate(date = vc)
    vc3.NextDay()

    if False:
        import io
        s = io.StringIO()
        clr = GCLocation.GCLocation()
        writeXml(s, clr, vc)
        GCUT.msg(s.getvalue())
        s.seek(0)
        writeGaurabdaTithiXml(s,clr,GCGaurabdaDate(2,1,512),GCGaurabdaDate(20,1,514))
        GCUT.msg('writeGaurabdaTithiXml')
        GCUT.msg(s.getvalue())

        s = io.StringIO()
        writeGaurabdaNextTithiXml(s,clr,vc,GCGaurabdaDate(2,1,512))
        GCUT.msg('writeGaurabdaNextTithiXml')
        GCUT.msg(s.getvalue())

    print(GetTithiTimes(e,vc,0.25))



from .GCEvent import GCEvent
from . import GCUT as GCUT
import os
import os.path
import json

glist_events = []

def get_list():
    if len(glist_events)==0:
        OpenFile('events.json')
    return glist_events

def add():
    c = GCEvent()
    glist_events.append(c)
    return c

def OpenFile(fileName):
    print('---------------- events loaded ----------------------')
    global glist_events
    if not os.path.exists(fileName):
        fileName = os.path.join(os.path.dirname(__file__), 'res', 'events.json')
    with open(fileName,'rt',encoding='utf-8') as rf:
        events = json.load(rf)
        for e in events:
            glist_events.append(GCEvent(data=e))
    return len(glist_events)

def SaveFile(fileName):
    with open(fileName,'wt',encoding='utf-8') as wf:
        events = [ce.data for ce in glist_events]
        wf.write(json.dumps(glist_events,indent=4))
    return len(glist_events)

def clear():
    glist_events = []

def SetOldStyleFasting(bOldStyle):
    locMatrix = []
    file_name = os.path.join(os.path.dirname(__file__), 'res', 'eventfast.json')
    with open(file_name,'rt',encoding='utf-8') as rf:
        locMatrix = json.load(rf)
    ret = 0
    key = 'fast' if bOldStyle else 'newfast'
    for a in locMatrix:
        for pce in glist_events:
            if pce.nMasa == a['masa'] and pce.nTithi == a['tithi'] and pce.nClass == a['cls']:
                if pce.nFastType != a[key]:
                    ret += 1
                    pce.nFastType = a[key]
                break
    return ret

def Count():
    return len(glist_events)

def EventAtIndex(index):
    return glist_events[index]


def unittests():
    GCUT.info('custom events')
    a = OpenFile('events.json')
    GCUT.nval(a,0,'open file')
    print(glist_events[0].data)



from .GCLocation import GCLocation
from . import GCCountry
from . import GCTimeZone
from . import GCUT

import os
import os.path
import json

locationList = []
m_bModified = True

def OpenFile(pszFileList):
    if not os.path.exists(pszFileList):
        pszFileList = os.path.join(os.path.dirname(__file__), 'res', 'locations.json')
    ImportFile(pszFileList,True)

def IsModified():
    return m_bModified

def SaveAs(lpszFileName):
    locations = [a.data() for a in locationList]
    with open(lpszFileName,'wt',encoding='utf-8') as json_file:
        json_file.write(json.dumps(locations))

def ImportFile(pszFile, bDeleteCurrent=False):
    global locationList
    with open(pszFile,'rt',encoding='utf-8') as json_file:
        locations = json.load(json_file)
        if bDeleteCurrent:
            locationList=[]
        for a in locations:
            locationList.append(GCLocation(data=a))
        m_bModified = True

def RemoveAll():
    locationList=[]
    m_bModified=True

def Add(loc):
    locationList.append(loc)
    m_bModified = True

def FindLocation(city=None,country=None):
    c = None if city is None else city.lower()
    cn = None if country is None else country.lower()
    for L in locationList:
        if c is None or c==L.m_strCity.lower():
            if cn is None or cn==L.m_strCountry.lower():
                return L
    return None

def FindLocations(city=None,country=None, limit=200):
    equals1 = []
    starting1 = []
    contains2 = []
    if city is not None:
        city = city.lower()
        # city is valid
        if country is not None:
            # country is valid
            print('1')
            country = country.lower()
            for L in locationList:
                if country != L.m_strCountry.lower():
                    continue
                c2 = L.m_strCity.lower()
                if c2 == city:
                    equals1.append(L)
                elif c2.startswith(city):
                    starting1.append(L)
                elif city in c2:
                    contains2.append(L)
                if len(equals1) + len(starting1) + len(contains2) >= limit:
                    return equals1, starting1, contains2
        else:
            # country is omitted
            print('2')
            for L in locationList:
                c2 = L.m_strCity.lower()
                if c2 == city:
                    equals1.append(L)
                elif c2.startswith(city):
                    starting1.append(L)
                elif city in c2:
                    contains2.append(L)
                if len(equals1) + len(starting1) + len(contains2) >= limit:
                    return equals1, starting1, contains2
    else:
        # city is omitted
        if country is not None:
            print('3')
            # country is valid
            country = country.lower()
            for L in locationList:
                c3 = L.m_strCountry.lower()
                if c3 == country:
                    equals1.append(L)
                elif c3.startswith(country):
                    starting1.append(L)
                elif country in c3:
                    contains2.append(L)
                if len(equals1) + len(starting1) + len(contains2) >= limit:
                    return equals1, starting1, contains2
        else:
            print('4')
            # country is omitted
            for L in locationList:
                equals1.append(L)
            if len(starting1) >= limit:
                return equals1, starting1, contains2
    return equals1, starting1, contains2

def GetLocationsForCountry(country=None,limit=-1):
    rv = []
    for L in locationList:
        if country==None or country==L.m_strCountry:
            if len(rv)<limit or limit<0:
                rv.append(L)
    return rv


def RemoveAt(index):
    del locationList[index]
    m_bModified = True

def RenameCountry(pszOld, pszNew):
    for loc in locationList:
        if loc.m_strCountry==pszOld:
            loc.m_strCountry = pszNew
            m_pModified = True

def LocationCount():
    return len(locationList)

def LocationAtIndex(index):
    return locationList[index]

def unittests():
    GCUT.info('location list')
    OpenFile('locations.json')
    GCUT.nval(LocationCount(),0,'location count')
    lc = LocationAtIndex(0)
    print(lc.data())



from .GCLocation import GCLocation
from .GCDayData import GCDayData
from .GCGregorianDate import GCGregorianDate, Today
from .GCStringBuilder import GCStringBuilder,SBTF_TEXT,SBTF_RTF
from math import floor
from . import GCUT as GCUT
import datetime
from . import GCMath as GCMath
from . import GCEarthData as GCEarthData
from . import GCAyanamsha as GCAyanamsha
from . import GCRasi as GCRasi
from .GCEnums.MasaId import *
from .GCPancangaDate import GCGaurabdaDate
from . import GCCalendar as GCCalendar
from . import GCStrings as GCStrings
from . import GCTimeZone as GCTimeZone
from . import GCDisplaySettings as GCDisplaySettings
from . import GCLayoutData as GCLayoutData

TRESULT_APP_CELEBS = 3


class TAppDay:
    def __init__(self):
        self.m_location = GCLocation()
        self.eventTime = GCGregorianDate()
        self.details = GCDayData()
        self.b_adhika = False
        self.celeb_gy = [0] * TRESULT_APP_CELEBS
        self.celeb_date = [GCGregorianDate() for i in range(TRESULT_APP_CELEBS)]

    def calculateAppDay(self, location, eventDate):
        d = self.details
        vc = GCGregorianDate(date = eventDate)
        vcsun = GCGregorianDate(date = eventDate)
        dprev = GCGregorianDate()
        dnext = GCGregorianDate()
        m_earth = location.GetEarthData()

        self.b_adhika = False
        self.eventTime.Set(eventDate)
        self.m_location.Set(location)

        vcsun.shour -= vcsun.tzone/24.0
        vcsun.NormalizeValues()
        vcsun.tzone = 0.0
        d.sun.SunPosition(vcsun, m_earth, vcsun.shour - 0.5)
        d.moon.Calculate(vcsun.GetJulianComplete(), m_earth)
        d.msDistance = GCMath.putIn360( d.moon.longitude_deg - d.sun.longitude_deg - 180.0)
        d.msAyanamsa = GCAyanamsha.GetAyanamsa( vc.GetJulianComplete() )

        # tithi
        dd = d.msDistance / 12.0
        d.nTithi = int(floor(dd))
        d.nTithiElapse = (dd - floor(dd)) * 100.0

        # naksatra
        dd = GCMath.putIn360( d.moon.longitude_deg - d.msAyanamsa )
        dd = ( dd * 3.0) / 40.0
        d.nNaksatra = int(floor(dd))
        d.nNaksatraElapse = (dd - floor(dd)) * 100.0
        d.nMasa = d.MasaCalc(vc, m_earth)
        d.nMoonRasi = GCRasi.GetRasi(d.moon.longitude_deg, d.msAyanamsa)
        d.nSunRasi = GCRasi.GetRasi(d.sun.longitude_deg, d.msAyanamsa)

        if (d.nMasa == ADHIKA_MASA):
            d.nMasa = GCRasi.GetRasi(d.sun.longitude_deg, d.msAyanamsa)
            self.b_adhika = True

        vc.Today()
        vc.tzone = m_earth.tzone
        m = 0
        va = GCGaurabdaDate()
        vctemp = GCGregorianDate()

        va.tithi = d.nTithi
        va.masa  = d.nMasa
        va.gyear = GCCalendar.GetGaurabdaYear(vc, m_earth)
        if (va.gyear < d.nGaurabdaYear):
            va.gyear = d.nGaurabdaYear


        for i in range(6):
            GCCalendar.Gaurabda2Gregorian(va, vctemp, m_earth)
            if (va.gyear > d.nGaurabdaYear):
                if (m < TRESULT_APP_CELEBS):
                    self.celeb_date[m].Set(vctemp)
                    self.celeb_gy[m] = va.gyear
                    m+=1
            va.gyear+=1

    def formatPlainText(self, stream):
        d = self.details
        vc = GCGregorianDate(date = self.eventTime)
        m_earth = self.m_location.GetEarthData()

        sb = GCStringBuilder(stream)
        sb.Format = SBTF_TEXT

        sb.AppendLine(GCStrings.getString(25))
        sb.AppendLine("")

        sb.AppendLine("{:15s} : {} {} {}".format(GCStrings.getString(7), vc.day, GCStrings.GetMonthAbreviation(vc.month), vc.year))

        sb.AppendLine("{:15s} : {}:{:02d}".format(GCStrings.getString(8), vc.GetHour(), vc.GetMinuteRound()))
        sb.AppendLine("")

        sb.AppendLine("{:15s} : {}".format(GCStrings.getString(9), self.m_location.m_strName))
        sb.AppendLine("{:15s} : {}".format(GCStrings.getString(10), self.m_location.m_strLatitude))
        sb.AppendLine("{:15s} : {}".format(GCStrings.getString(11), self.m_location.m_strLongitude))
        sb.AppendLine("{:15s} : {}".format(GCStrings.getString(12),  self.m_location.m_strTimeZone))
        sb.AppendLine("{:15s} : N/A".format("DST"))
        sb.AppendLine("")

        sb.AppendLine("{:15s} : {}".format(GCStrings.getString(13), GCStrings.GetTithiName(d.nTithi)))
        sb.AppendLine("{:15s} : {:.2f} %".format(GCStrings.getString(14), d.nTithiElapse))
        sb.AppendLine("{:15s} : {}".format(GCStrings.getString(15), GCStrings.GetNaksatraName(d.nNaksatra)))
        sb.AppendLine("{:15s} : {:.2f} % ({} pada)".format(GCStrings.getString(16), d.nNaksatraElapse,
            GCStrings.getString(811+int(d.nNaksatraElapse/25.0))))
        sb.AppendLine("{:15s} : {} ({})".format("Moon Rasi", GCStrings.GetSankrantiName(d.nMoonRasi), GCStrings.GetSankrantiNameEn(d.nMoonRasi)))
        sb.AppendLine("{:15s} : {} ({})".format("Sun Rasi", GCStrings.GetSankrantiName(d.nSunRasi), GCStrings.GetSankrantiNameEn(d.nSunRasi)))

        sb.AppendLine("{:15s} : {}".format(GCStrings.getString(20), GCStrings.GetPaksaName(d.nPaksa)))
        if (self.b_adhika == True):
            sb.AppendLine("{:15s} : {} {}".format(GCStrings.getString(22), GCStrings.GetMasaName(d.nMasa), GCStrings.getString(21)))
        else:
            sb.AppendLine("{:15s} : {}".format(GCStrings.getString(22), GCStrings.GetMasaName(d.nMasa)))
        sb.AppendLine("{:15s} : {}".format(GCStrings.getString(23), d.nGaurabdaYear))

        if (GCDisplaySettings.getValue(48)):
            sb.AppendLine()
            sb.AppendLine(GCStrings.getString(17))
            sb.AppendLine()

            sb.AppendLine("%25s : {}... ".format(GCStrings.getString(18) , GCStrings.GetNaksatraChildSylable(d.nNaksatra, int(d.nNaksatraElapse/25.0))))
            sb.AppendLine("%25s : {}... ".format(GCStrings.getString(19), GCStrings.GetRasiChildSylable(d.nMoonRasi)))

        sb.AppendLine()
        sb.AppendLine(GCStrings.getString(24))
        sb.AppendLine()

        for o in range(TRESULT_APP_CELEBS):
            sb.AppendLine("   Gaurabda {:3d} : {} ".format(self.celeb_gy[o] , self.celeb_date[o]))

        sb.AppendNote()

    def formatRtf(self,stream):
        d = self.details
        vc = GCGregorianDate(date=self.eventTime)
        m_earth = self.m_location.GetEarthData()

        sb = GCStringBuilder(stream)
        sb.fontSizeH1 = GCLayoutData.textSizeH1
        sb.fontSizeH2 = GCLayoutData.textSizeH2
        sb.fontSizeText = GCLayoutData.textSizeText
        sb.fontSizeNote = GCLayoutData.textSizeNote
        sb.Format = SBTF_RTF

        sb.AppendDocumentHeader()

        sb.AppendHeader1(GCStrings.getString(25))


        sb.AppendLine("\\tab {} : {{\\b {} {} {} }}".format(GCStrings.getString(7), vc.day, GCStrings.GetMonthAbreviation(vc.month), vc.year))

        sb.AppendLine("\\tab {} : {{\\b {}:{:02d} }}".format(GCStrings.getString(8), vc.GetHour(), vc.GetMinuteRound()))
        sb.AppendLine("")

        sb.AppendLine("\\tab {} : {{\\b {} }}".format(GCStrings.getString(9), self.m_location.m_strName))
        sb.AppendLine("\\tab {} : {{\\b {} }}".format(GCStrings.getString(10), self.m_location.m_strLatitude))
        sb.AppendLine("\\tab {} : {{\\b {} }}".format(GCStrings.getString(11), self.m_location.m_strLongitude))
        sb.AppendLine("\\tab {} : {{\\b {} }}".format(GCStrings.getString(12), GCTimeZone.GetTimeZoneName(self.m_location.m_nTimezoneId)))
        sb.AppendLine("\\tab {} : {{\\b N/A }}".format("DST"))
        sb.AppendLine("")

        sb.AppendLine("\\tab {} : {{\\b {} }}".format(GCStrings.getString(13), GCStrings.GetTithiName(d.nTithi)))
        sb.AppendLine("\\tab {} : {{\\b {:.2f} % }}".format(GCStrings.getString(14), d.nTithiElapse))
        sb.AppendLine("\\tab {} : {{\\b {} }}".format(GCStrings.getString(15), GCStrings.GetNaksatraName(d.nNaksatra)))
        sb.AppendLine("\\tab {} : {{\\b {:.2f} % ({} pada) }}".format(GCStrings.getString(16), d.nNaksatraElapse, GCStrings.getString(811+int(d.nNaksatraElapse/25.0))))
        sb.AppendLine("\\tab {} : {{\\b {} ({}) }}".format("Moon Rasi", GCStrings.GetSankrantiName(d.nMoonRasi), GCStrings.GetSankrantiNameEn(d.nMoonRasi)))
        sb.AppendLine("\\tab {} : {{\\b {} ({}) }}".format("Sun Rasi", GCStrings.GetSankrantiName(d.nSunRasi), GCStrings.GetSankrantiNameEn(d.nSunRasi)))

        sb.AppendLine("\\tab {} : {{\\b {} }}".format(GCStrings.getString(20), GCStrings.GetPaksaName(d.nPaksa)))
        if (self.b_adhika == True):
            sb.AppendLine("\\tab {} : {{\\b {} {} }}".format(GCStrings.getString(22), GCStrings.GetMasaName(d.nMasa), GCStrings.getString(21)))
        else:
            sb.AppendLine("\\tab {} : {{\\b {} }}".format(GCStrings.getString(22), GCStrings.GetMasaName(d.nMasa)))
        sb.AppendLine("\\tab {} : {{\\b {} }}".format(GCStrings.getString(23), d.nGaurabdaYear))

        if (GCDisplaySettings.getValue(48)):
            sb.AppendLine("")
            sb.AppendHeader2(GCStrings.getString(17))
            sb.AppendLine("")
            sb.AppendLine("\\tab {} : {{\\b {}...  }}".format(GCStrings.getString(18) , GCStrings.GetNaksatraChildSylable(d.nNaksatra, int(d.nNaksatraElapse/25.0))))
            sb.AppendLine("\\tab {} : {{\\b {}...  }}".format(GCStrings.getString(19), GCStrings.GetRasiChildSylable(d.nMoonRasi)))


        sb.AppendLine("{{\\fs{} {} }}".format(GCLayoutData.textSizeH2, GCStrings.getString(24)))
        sb.AppendLine("")

        for o in range(TRESULT_APP_CELEBS):
            sb.AppendLine("\\tab Gaurabda {} : {{\\b {} }}".format(self.celeb_gy[o] , self.celeb_date[o]))

        sb.AppendDocumentTail()

    def formatXml(self,stream):
        d = self.details
        vc = GCGregorianDate(date=self.eventTime)
        m_earth = self.m_location.GetEarthData()
        loc = self.m_location
        bDuringAdhika = False

        stream.write('''<xml>
    <request name="AppDay" version="{}">
        <arg name="longitude" value="{}" />
        <arg name="latitude" value="{}" />
        <arg name="timezone" value="{}" />
        <arg name="year" value="{}" />
        <arg name="month" value="{}" />
        <arg name="day" value="{}" />
        <arg name="hour" value="{}" />
        <arg name="minute" value="{}" />
    </request>'''.format(GCStrings.getString(130), loc.m_fLongitude, loc.m_fLatitude, loc.m_fTimezone, self.eventTime.year, self.eventTime.month, self.eventTime.day, self.eventTime.GetHour(), self.eventTime.GetMinuteRound() ))

        npada = int (floor(d.nNaksatraElapse / 25.0)) + 1
        if (npada > 4): npada = 4
        is_adhika = "yes" if bDuringAdhika else "no"
        stream.write("\t<result name=\"AppDay\" >\n\t\t<tithi name=\"{}\" elapse=\"%f\" />\n\t\t<naksatra name=\"{}\" elapse=\"%f\" pada=\"{}\"/>\n\t\t<paksa name=\"{}\" />\n\t\t<masa name=\"{}\" adhikamasa=\"{}\"/>\n\t\t<gaurabda value=\"{}\" />\n".format(GCStrings.GetTithiName(d.nTithi), d.nTithiElapse , GCStrings.GetNaksatraName(d.nNaksatra), d.nNaksatraElapse, npada , GCStrings.GetPaksaName(d.nPaksa) , GCStrings.GetMasaName(d.nMasa), is_adhika , d.nGaurabdaYear ))

        stream.write("\t\t<celebrations>\n")
        for i in range(TRESULT_APP_CELEBS):
            stream.write("\t\t\t<celebration gaurabda=\"{}\" day=\"{}\" month=\"{}\" monthabr=\"{}\" year=\"{}\" />\n".format(self.celeb_gy[i], self.celeb_date[i].day, self.celeb_date[i].month, GCStrings.GetMonthAbreviation(self.celeb_date[i].month), self.celeb_date[i].year))

        stream.write("\t\t</celebrations>\n\t</result>\n</xml>\n")


    def writeHtml(self,stream):
        d = self.details
        vc = GCGregorianDate(date = self.eventTime)
        m_earth = self.m_location.GetEarthData()

        stream.write("<html><head><title>Appearance day</title>")
        stream.write("<style>\n<!--\nbody {\n  font-family:Verdana;\n  font-size:11pt;\n}\n\ntd.hed {\n  font-size:11pt;\n  font-weight:bold;\n")
        stream.write("  background:#aaaaaa;\n  color:white;\n  text-align:center;\n  vertical-align:center;\n  padding-left:15pt;\n  padding-right:15pt;\n")
        stream.write("  padding-top:5pt;\n  padding-bottom:5pt;\n}\n-->\n</style>\n")
        stream.write("</head>\n\n<body>\n")
        stream.write("<h2 align=center>Appearance day Calculation</h2>")
        stream.write("<table align=center><tr><td valign=top>\n\n")
        stream.write("<table align=center>")
        stream.write("<tr><td colspan=3 class=hed>Details</td></tr>\n")
        stream.write("<tr><td colspan=2>{}</td><td> {}</td></tr>\n".format(GCStrings.getString(7), vc))
        stream.write("<tr><td colspan=2>{}</td><td> {}:{:02d}</td></tr>\n\n".format(GCStrings.getString(8), vc.GetHour(), vc.GetMinuteRound()))
        stream.write("<tr><td colspan=2>{}</td><td> {}</td></tr>\n".format(GCStrings.getString(9), self.m_location.m_strName))
        stream.write("<tr><td colspan=2>{}</td><td> {}</td></tr>\n".format(GCStrings.getString(10), GCEarthData.GetTextLatitude(self.m_location.m_fLatitude)))
        stream.write("<tr><td colspan=2>{}</td><td> {}</td></tr>\n".format(GCStrings.getString(11), GCEarthData.GetTextLongitude(self.m_location.m_fLongitude)))
        stream.write("<tr><td colspan=2>{}</td><td> ".format(GCStrings.getString(12)))
        stream.write(GCTimeZone.GetTimeZoneOffsetText(self.m_location.m_fTimezone))
        stream.write("</td></tr>\n")
        stream.write("<tr><td colspan=2>DST</td><td>N/A</td></tr>\n")
        stream.write("<tr><td colspan=2>{}</td><td> {}</td></tr>\n".format(GCStrings.getString(13), GCStrings.GetTithiName(d.nTithi)))
        stream.write("<tr><td colspan=2>{}</td><td> {:.2f} %</td></tr>\n".format(GCStrings.getString(14), d.nTithiElapse))
        stream.write("<tr><td colspan=2>{}</td><td> {}</td></tr>\n".format(GCStrings.getString(15), GCStrings.GetNaksatraName(d.nNaksatra)))
        stream.write("<tr><td colspan=2>{}</td><td> {:.2f} %</td></tr>\n".format(GCStrings.getString(16), d.nNaksatraElapse))
        stream.write("<tr><td colspan=2>{}</td><td> {}</td></tr>\n".format(GCStrings.getString(20), GCStrings.GetPaksaName(d.nPaksa)))
        if (self.b_adhika == True):
            stream.write("<tr><td colspan=2>{}</td><td> {} {}</td></tr>\n".format(GCStrings.getString(22), GCStrings.GetMasaName(d.nMasa), GCStrings.getString(21)))
        else:
            stream.write("<tr><td colspan=2>{}</td><td> {}</td></tr>\n".format(GCStrings.getString(22), GCStrings.GetMasaName(d.nMasa)))
        stream.write("<tr><td colspan=2>{}</td><td> {}</td></tr>\n\n".format(GCStrings.getString(23), d.nGaurabdaYear))

        stream.write("</table></td><td valign=top><table>")
        stream.write("<tr><td colspan=3 class=hed>{}</td></tr>\n".format(GCStrings.getString(24)))

        for o in range(TRESULT_APP_CELEBS):
            stream.write("<tr><td>Gaurabda {}</td><td>&nbsp;&nbsp;:&nbsp;&nbsp;</td><td><b>{}</b></td></tr>".format(self.celeb_gy[o] , self.celeb_date[o]))
        stream.write("</table>")
        stream.write("</td></tr></table>\n\n")
        stream.write("<hr align=center width=\"50%\">\n<p style=\'text-align:center;font-size:8pt\'>Generated by {}</p>".format(GCStrings.getString(130)))
        stream.write("</body></html>")

    def write(self,stream,format='html',layout='list'):
        if format=='plain':
            self.formatPlainText(stream)
        elif format=='rtf':
            self.formatRtf(stream)
        elif format=='xml':
            self.formatXml(stream)
        elif format=='html':
            self.writeHtml(stream)

def unittests():
    GCUT.info('result app')
    loc = GCLocation(data={
        'latitude': 48.150002,
        'longitude': 17.116667,
        'tzid': 321,
        'name': 'Bratislava, Slovakia'
    })
    earth = loc.GetEarthData()
    today = Today()

    tc = TAppDay()

    print('start calculate', datetime.datetime.now())
    tc.calculateAppDay(loc,today)
    print('end calculate', datetime.datetime.now())

    with open('test/app.xml','wt') as wf:
        tc.formatXml(wf)
    with open('test/app.txt','wt') as wf:
        tc.formatPlainText(wf)
    with open('test/app.rtf','wt') as wf:
        tc.formatRtf(wf)
    with open('test/app.html','wt') as wf:
        tc.writeHtml(wf)



import math
import datetime
import re

from . import GCMath as GCMath
from . import GCUT as GCUT
from . import GCStrings as GCStrings

def GetMonthMaxDays(year, month):
    m_months = [ 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 ]

    if IsLeapYear(year) and month==2: return 29
    return m_months[month]

def IsLeapYear(year):
    if (year % 4) != 0: return False
    if year % 100 == 0 and year % 400 != 0: return False
    return True

def NormalizedValues(date, tzone=False):
    d1 = date.day
    m1 = date.month
    y1 = date.year
    h1 = date.shour
    if tzone: h1 += date.tzone/24.0

    if h1 < 0.0:
        d1-=1
        h1 += 1.0
    elif h1 >= 1.0:
        h1 -= 1.0
        d1+=1

    if d1 < 1:
        m1-=1
        if m1 < 1:
            m1 = 12
            y1-=1
        d1 = GetMonthMaxDays(y1, m1)
    elif d1 > GetMonthMaxDays(y1, m1):
        m1+=1
        if m1 > 12:
            m1 = 1
            y1+=1
        d1 = 1
    return y1,m1,d1,h1

def CalculateJulianDay(year, month, day):
    yy = year - int((12 - month) / 10)
    mm = month + 9

    if mm >= 12:
        mm -= 12

    k1 = int (math.floor(365.25 * (yy + 4712)))
    k2 = int (math.floor(30.6 * mm + 0.5))
    k3 = int (math.floor(math.floor((yy/100.0)+49)*.75))-38
    j = k1 + k2 + day + 59
    if j > 2299160:
        j -= k3

    return j


class GCGregorianDate:
    def __init__(self,year=None,month=None,day=None,shour=None,tzone=None,date=None,text=None,addDays=0):
        self.year = 2000
        self.month = 1
        self.day = 1
        self.shour = 0.5
        self.tzone = 0.0
        if text != None: self.SetText(text)
        if date!=None: self.Set(date)
        if year != None: self.year = year
        if month != None: self.month = month
        if day != None: self.day = day
        if shour != None: self.shour = shour
        if tzone != None: self.tzone = tzone
        self.InitWeekDay()
        if addDays!=0:
            self.AddDays(addDays)

    def copy(self):
        return GCGregorianDate(date=self)

    def __str__(self):
        return '{} {} {:04d}'.format(self.day, GCStrings.GetMonthAbreviation(self.month), self.year)

    def __repr__(self):
        return "{:2d} {} {:04d}  {:02d}:{:02d}:{:02d}".format(self.day, GCStrings.GetMonthAbreviation(self.month), self.year, self.GetHour(), self.GetMinute(), self.GetSecond())

    def __dict__(self):
        return {
            'year': self.year,
            'month': self.month,
            'day': self.day,
            'hour': self.GetHour(),
            'minute': self.GetMinute(),
            'second': self.GetSecond(),
            'offset': self.tzone
        }

    def __iter__(self):
        yield 'year', self.year,
        yield 'month', self.month,
        yield 'day', self.day,
        yield 'hour', self.GetHour(),
        yield 'minute', self.GetMinute(),
        yield 'second', self.GetSecond(),
        yield 'offset', self.tzone


    def time_str(self):
        return "{:02d}:{:02d}:{:02d}".format(self.GetHour(), self.GetMinute(), self.GetSecond())

    def MonthAbrToInt(self,text):
        if len(text)>3: text=text[:3]
        for i in range(65,78):
            if GCStrings.getString(i).lower()==text.lower():
                return i-64
        raise

    def SetText(self,text):
        y = 2000
        m = 1
        d = 1
        sh = 0.5
        tz = 0.0
        mr = re.match( r'([0-9]+) ([A-Za-z]+) ([0-9]+)', text )
        if mr:
            try:
                d = int(mr.group(1))
                m = self.MonthAbrToInt(mr.group(2))
                y = int(mr.group(3))
                self.year = y
                self.month = m
                self.day = d
                self.NormalizeValues()
            except:
                return
            return


    def IsLessThan(self,date):
        y1,m1,d1,h1 = NormalizedValues(self,tzone=True)
        y2,m2,d2,h2 = NormalizedValues(date,tzone=True)

        if y1 > y2: return False
        if y1 < y2: return True

        if m1 > m2: return False
        if m1 < m2: return True

        if d1 < d2: return True
        return False

    def IsLessOrEqualTo(self,date):
        y1,m1,d1,h1 = NormalizedValues(self,tzone=True)
        y2,m2,d2,h2 = NormalizedValues(date,tzone=True)

        if y1 > y2: return False
        if y1 < y2: return True

        if m1 > m2: return False
        if m1 < m2: return True

        if d1 <= d2: return True
        return False

    def IsDateEqual(self,date):
        y1,m1,d1,h1 = NormalizedValues(self,tzone=True)
        y2,m2,d2,h2 = NormalizedValues(date,tzone=True)

        return y1 == y2 and m1 == m2 and d1 == d2

    def NormalizeValues(self):
        self.year,self.month,self.day,self.shour = NormalizedValues(self)

    def NormalizeHours(self):
        while self.shour < 0.0:
            self.shour += 1.0
            self.PreviousDay()
        while self.shour > 1.0:
            self.shour -= 1.0
            self.NextDay()

    def ChangeTimeZone(self,tZone):
        self.shour += (tZone - self.tzone)/24
        self.NormalizeValues()
        self.tzone = tZone

    def GetJulianInteger(self):
        return CalculateJulianDay(self.year,self.month,self.day)

    def GetJulian(self):
        return self.GetJulianInteger() * 1.0

    def PreviousDay(self):
        self.day-=1
        if self.day < 1:
            self.month-=1
            if self.month < 1:
                self.month = 12
                self.year-=1
            self.day = GetMonthMaxDays(self.year, self.month)
        self.dayOfWeek = (self.dayOfWeek + 6) % 7

    def NextDay(self):
        self.day+=1
        if self.day > GetMonthMaxDays(self.year, self.month):
            self.month+=1
            if self.month > 12:
                self.month = 1
                self.year+=1
            self.day = 1
        self.dayOfWeek = (self.dayOfWeek + 1) % 7

    def AddDays(self,n):
        if n>0:
            for i in range(n):
                self.NextDay()
        elif n<0:
            for i in range(-n):
                self.PreviousDay()

    def SubtractDays(self, n):
        self.AddDays(-n)

    def GetJulianDetailed(self):
        return self.GetJulian() - 0.5 + self.shour

    def GetJulianComplete(self):
        return self.GetJulian() - 0.5 + self.shour - self.tzone/24.0

    def InitWeekDay(self):
        self.dayOfWeek = (int(self.GetJulianInteger()) + 1) % 7;

    def SetFromJulian(self,jd):
        z = math.floor(jd + 0.5)
        f = (jd + 0.5) - z
        if z < 2299161.0:
            A = z
        else:
            alpha = math.floor((z - 1867216.25)/36524.25)
            A = z + 1.0 + alpha - math.floor(alpha/4.0)

        B = A + 1524
        C = math.floor((B - 122.1)/365.25)
        D = math.floor(365.25 * C)
        E = math.floor((B - D)/30.6001)
        self.day = int(math.floor(B - D - floor(30.6001 * E) + f))
        if E < 14:
            self.month = int(E - 1)
        else:
            self.month = int(E - 13)
        if self.month > 2:
            self.year = int(C - 4716)
        else:
            self.year = int(C - 4715)
        self.tzone = 0.0
        self.shour = GCMath.getFraction(jd + 0.5)

    def Clear(self):
        self.year = 0
        self.month = 0
        self.day = 0
        self.shour = 0.0
        self.tzone = 0.0

    def Set(self,ta):
        self.year = ta.year
        self.month = ta.month
        self.day = ta.day
        self.shour = ta.shour
        self.tzone = ta.tzone
        self.dayOfWeek = ta.dayOfWeek

    def IsBeforeThis(self,date):
        if self.year > date.year: return False
        if self.year < date.year: return True

        if self.month > date.month: return False
        if self.month < date.month: return True

        return self.day < date.day

    def Today(self):
        d = datetime.datetime.now()
        self.year = d.year
        self.month = d.month
        self.day = d.day
        self.shour = 0.5
        self.tzone = 0
        self.dayOfWeek = d.weekday()

    def CompareYMD(self,v):
        if v.year < self.year:
            return (self.year - v.year) * 365
        if v.year > self.year:
            return (self.year - v.year) * 365;

        if v.month < self.month:
            return (self.month - v.month)*31
        if v.month > self.month:
            return (self.month - v.month)*31

        return self.day - v.day

    def GetHour(self):
        return int(math.floor(self.shour*24))

    def GetMinute(self):
        return int(math.floor(GCMath.getFraction(self.shour*24) * 60))

    def GetMinuteRound(self):
        return int(math.floor(GCMath.getFraction(self.shour*24) * 60 + 0.5))

    def GetDayInteger(self):
        return self.year * 384 + self.month * 32 + self.day

    def GetSecond(self):
        return int(math.floor(GCMath.getFraction(self.shour*1440) * 60))

    def GetDateTextWithTodayExt(self):
        if (self.day > 0) and (self.day < 32) and (self.month > 0) \
            and (self.month < 13) and (self.year >= 1500) and (self.year < 4000):
            today = GCGregorianDate()
            diff = today.GetJulianInteger() - self.GetJulianInteger()
            if diff == 0:
                return str(self) + ' ' + GCStrings.getString(43)
            elif diff == -1:
                return str(self) + ' ' + GCStrings.getString(854)
            elif diff == 1:
                return str(self) + ' ' + GCStrings.getString(853)
            else:
                return str(self)

def Today():
    vc = GCGregorianDate()
    vc.Today()
    return vc

def unittests():
    GCUT.info('GCGregorianDate')
    t = GCGregorianDate()
    GCUT.val(t.day,0,'zero day')
    t.Today()
    GCUT.nval(t.day,0,'init day')
    GCUT.nval(t.month,0,'init month')
    GCUT.msg('Today: ' + str(t))
    GCUT.msg('Now: ' + repr(t))
